#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

''' ESP Windows '''

from __future__ import print_function, division, absolute_import, unicode_literals

import collections
import copy
import csv
import datetime
import functools
import inspect
import itertools
import os
import pandas as pd
import re
import requests
import six
import sys
import threading
import types
import weakref
import xml.etree.ElementTree as ET
from six.moves import urllib
from .features import (WindowFeature, SplitterExpressionFeature,
                       SplitterPluginFeature, FinalizedCallbackFeature,
                       ConnectorsFeature)
from .subscriber import Subscriber
from .publisher import Publisher
from .utils import listify, get_args, ensure_element, connectors_to_end
from .. import transformers
from ..base import ESPObject, attribute
from ..config import get_option, ESP_ROOT, CONCAT_OPTIONS
from ..exceptions import ESPError
from ..plotting import StreamingChart, StreamingImages, split_chart_params
from ..schema import Schema
from ..utils.keyword import dekeywordify
from ..utils import xml
from ..utils.notebook import scale_svg
from ..utils.rest import get_params
from ..utils.data import get_project_data, gen_name, get_server_info
from ..utils.events import get_events, get_dataframe, get_schema
from ..websocket import WebSocketClient

INDEX_TYPES = {
    'rbtree': 'pi_RBTREE',
    'hash': 'pi_HASH',
    'ln_hash': 'pi_LN_HASH',
    'cl_hash': 'pi_CL_HASH',
    'fw_hash': 'pi_FW_HASH',
    'empty': 'pi_EMPTY'
}


def get_window_class(name):
    '''
    Return the class for the specified window name

    Parameters
    ----------
    name : string
        The name of the window type

    Returns
    -------
    :class:`BaseWindow` subclass

    '''
    if not name.startswith('window-'):
        name = 'window-%s' % name
    return BaseWindow.window_classes[name]


def var_mapper(data, mapping):
    '''
    Map data column values to generated values

    Parameters
    ----------
    data : DataFrame
        The DataFrame to access data from
    mapping : dict
        A nested dictionary where the top-level keys are the column
        names.  The values are a dictionary where the keys are
        the variable to generate and the values are dictionaries
        with the old-value / new-value pairs.

    Returns
    -------
    dict of lists

    '''
    out = {}
    for col, attrs in mapping.items():
        for aname, attrmap in attrs.items():
            if isinstance(attrmap, dict):
                out[aname] = list(data[col].map(attrmap))
            else:
                out[aname] = list(attrmap(data[col]))
    return out


def param_iter(params):
    '''
    Iterate over all combinations of parameters

    Parameters
    ----------
    params : dict or list of dicts
        The sets of parameters

    Examples
    --------
    >>> for item in param_iter(dict(a=1, b=['x', 'y', 'z'], c=100)):
    ...      print(item)
    ...      if item['b'] == 'z':
    ...          break
    {'a': 1, b: 'x', c: 100}
    {'a': 1, b: 'y', c: 100}
    {'a': 1, b: 'z', c: 100}

    >>> for item in param_iter([dict(a=1, b='x', c=100),
    ...                         dict(a=1, b='y'),
    ...                         dict(a=2, c=200)]):
    ...    print(item)
    ...    if item['a'] == 2:
    ...        break
    {'a': 1, 'b': 'x', 'c': 100}
    {'a': 1, 'b': 'y'}
    {'a': 2, 'c': 200}

    Yields
    ------
    dict

    '''
    while True:
        if not params:
            yield {}
        elif isinstance(params, collections.Mapping):
            keys, values = zip(*params.items())
            values = list(values)
            n_items = 1
            for val in values:
                if isinstance(val, (tuple, list, set)):
                    n_items = max(n_items, len(val))
            for i, val in enumerate(values):
                if isinstance(val, six.string_types) or \
                        not isinstance(val, collections.Sequence):
                    values[i] = [val] * n_items
            for i in range(n_items):
                out = {}
                for key, value in zip(keys, values):
                    out[key] = value[i]
                yield out
        else:
            for value in params:
                yield copy.copy(value)


class Target(object):
    '''
    Window target

    Parameters
    ----------
    name : string
        The name of the target window
    role : string, optional
        The role of the connection

    Returns
    -------
    :class:`Target`

    '''
    _index = 0
    _index_lock = threading.Lock()

    def __init__(self, name, role=None, slot=None):
        self.name = name
        self.role = role
        self.slot = slot
        self._index = self._get_new_index()

    @classmethod
    def _get_new_index(cls):
        with cls._index_lock:
            cls._index += 1
            return cls._index

    def __hash__(self):
        return id(self)

    def __eq__(self, other):
        return hash(other) == hash(self)

    def __str__(self):
        if self.role is None and self.slot is None:
            return '%s(%s)' % (type(self).__name__, repr(self.name))
        if self.role is not None and self.slot is not None:
            return '%s(%s, role=%s, slot=%s)' % (type(self).__name__, repr(self.name),
                                                 repr(self.role), repr(self.slot))
        if self.role is not None:
            return '%s(%s, role=%s)' % (type(self).__name__, repr(self.name), repr(self.role))
        return '%s(%s, slot=%s)' % (type(self).__name__, repr(self.name), repr(self.slot))

    def __repr__(self):
        return str(self)


class BaseWindow(ESPObject):
    '''
    Base Window class

    Parameters
    ----------
    name : string, optional
        The name of the window

    Attributes
    ----------
    contquery : string
        The name of the continuous query the window is associated with
    data : pandas.DataFrame
        The cached rows of event data
    description : string
        A description of the window
    event_transformers : list
        A list of event transformer definitions
    project : string
        The name of the project the window is associated with
    targets : set
        Set of target definitions
    url : string
        The URL of the window

    Notes
    -----
    All constructor parameters are also available as attributes.

    Returns
    -------
    :class:`Window`

    '''

    window_type = None
    window_classes = {}

    name = attribute('name', dtype='string')
    pubsub = attribute('pubsub', dtype='bool')

    def __init__(self, name=None, **kwargs):
        schema = kwargs.pop('schema', None)
        copyvars = kwargs.pop('copyvars', None)
        ESPObject.__init__(self, attrs=kwargs)
        self.schema = Schema()
        if schema is not None:
            self.schema = schema
        self.name = name or gen_name(prefix='w_')
        self.contquery = None
        self.project = None
        self.targets = set()
        self.data = None
        self.description = None
        self.event_transformers = []
        self._subscriber = None
        self._initialize_features()
        self.copyvars = list(copyvars or [])

    def _initialize_features(self):
        features = WindowFeature.__subclasses__()
        for item in inspect.getmro(type(self)):
            if item in features:
                item.__init__(self)

    def _cast_parameter(self, value, dtype=None, vartype=None):
        '''
        Cast parameter to the appropriate type

        Parameters
        ----------
        value : any
            The value of the window parameter
        dtype : string, optional
            The data type of the parameter
        vartype : string, optional
            The sub-data type of the parameter.  Typically used when
            `dtype` is a list parameter type.

        Returns
        -------
        any

        '''
        if dtype == 'varlist':
            if isinstance(value, six.string_types):
                value = re.split(r'\s*,\s*', value.strip())
        return value

    def _export_copyvars(self, out=None):
        ''' Get schema for copyvars '''
        def get_field_parts(field, map_name, dtypes):
            key = False
            dtype = 'double'
            if '*' in field:
                field = field.replace('*', '')
                key = True
            if ':' in field:
                field, dtype = field.split(':', 1)
            elif map_name in dtypes:
                dtype = dtypes[map_name]
            # Remove array ranges
            if re.search(r'\[[^\]]+\]$', field):
                field = re.sub(r'\[[^\]]+\]$', r'', field)
                dtype = 'array(%s)' % dtype
                dtype = dtype.replace('double', 'dbl')
                dtype = dtype.replace('int64', 'i64')
                dtype = dtype.replace('int32', 'i32')
            return field, dtype, key

        if out is None:
            out = self._schema.copy()

        online = getattr(self, 'online_models', [])
        offline = getattr(self, 'offline_models', [])

        maps = []
        input_map_types = getattr(type(self), 'input_map_types', {})
        output_map_types = getattr(type(self), 'output_map_types', {})

        if type(self).window_type != 'train':
            maps.append([getattr(self, 'input_map', {}), input_map_types])
            for model in online:
                maps.append([getattr(model, 'input_map', {}), input_map_types])
            for model in offline:
                maps.append([getattr(model, 'input_map', {}), input_map_types])

        maps.append([getattr(self, 'output_map', {}), output_map_types])
        for model in online:
            maps.append([getattr(model, 'output_map', {}), output_map_types])
        for model in offline:
            maps.append([getattr(model, 'output_map', {}), output_map_types])

        copied_fields = {}
        for item, dtypes in maps:
            for map_name, value in item.items():
#               if 'id' not in out:
#                   out['id'] = 'int64', True
                for field in listify(value or []):
                    name, dtype, is_key = get_field_parts(field, map_name, dtypes)
                    for part in re.split(r'\s*,\s*', name.strip()):
                        if part in self.copyvars:
                            copied_fields[part] = dtype, is_key

        for item in self.copyvars:
            if item in copied_fields:
                out[item] = copied_fields[item]

        return out

    @property
    def schema(self):
        '''
        Return the schema for the window

        Returns
        -------
        :class:`Schema`

        '''
        if not self._schema.fields or self.copyvars:
            # Build a transient schema based on input and output maps
            out = self._schema.copy()
            out = self._export_copyvars(out)
            return out

        return self._schema

    @schema.setter
    def schema(self, value):
        '''
        Set a new window schema

        Parameters
        ----------
        value : Schema or string or tuple
            The value to set the schema to.  If it is a string, it should
            be a valid schema string for an entire window.  If it is a
            tuple, each component of the tuple should be a schema string
            for that column.

        '''
        if isinstance(value, six.string_types):
            self._schema.schema_string = value
        elif isinstance(value, Schema):
            self._schema = value
        elif isinstance(value, (list, tuple)):
            self._schema.schema_string = ' '.join(value)
        else:
            raise TypeError('Unknown type for schema: %s' % value)

    @property
    def schema_string(self):
        '''
        The window schema as a string representation

        Returns
        -------
        string

        '''
        return self.schema.schema_string

    @schema_string.setter
    def schema_string(self, value):
        '''
        Set the window schema from a schema string

        Parameters
        ----------
        value : string
            The schema string to use for the window schema

        '''
        self._schema.schema_string = value

    @property
    def subscriber_url(self):
        '''
        Return the subscriber URL for web sockets

        Returns
        -------
        string

        '''
        return re.sub(r'^\w+(:.+?/%s/)' % ESP_ROOT, r'ws\1subscribers/', self.url)

    @property
    def publisher_url(self):
        '''
        Return the publisher URL for web sockets

        Returns
        -------
        string

        '''
        return re.sub(r'^\w+(:.+?/%s/)' % ESP_ROOT, r'ws\1publishers/', self.url)

    def add_event_transformer(self, method, *args, **kwargs):
        '''
        Add an transformer to apply for each event

        Parameters
        ----------
        method : string or callable
            If a string, it is the name of a :class:`pandas.DataFrame`
            method to execute.  This method will be executed on the
            returned event DataFrame before the result is appended to
            the Window's DataFrame.
            If a callable, the first argument must be the event DataFrame,
            and the new DataFrame must be returned.
        *args : one-or-more arguments, optional
            Arguments to pass to the transformer
        **kwargs : keyword arguments, optional
            Keyword arguments to pass to the transformer

        Examples
        --------
        Add DataFrame method transformer.

        >>> win.add_event_transformer('clip', upper=20)

        Add an arbitrary function transformer.

        >>> def new_col(data):
        ...     data['new_col'] = (data['a'] + data['b'])**2
        ...     return data
        ...
        >>> win.add_event_transfromer(new_col)

        '''
        self.event_transformers.append((method, tuple(args), dict(kwargs)))
        if self.data is not None:
            if isinstance(method, six.string_types):
                if method in ['rgb2bgr', 'bgr2rgb', 'bytes2image']:
                    self.data = getattr(transformers, method)(data, *tuple(args), **dict(kwargs))
                else:
                    self.data = getattr(self.data, method)(*tuple(args), **dict(kwargs))
            else:
                self.data = method(self.data, *tuple(args), **dict(kwargs))

    def create_subscriber(self, mode='updating', pagesize=50, filter=None,
                          sort=None, format='xml', separator=None,
                          interval=None, schema=False,
                          on_event=None, on_message=None, on_error=None,
                          on_close=None, on_open=None):
        '''
        Create a new websocket subscriber for the window

        Parameters
        ----------
        mode : string, optional
            The type of subscriber: 'updating' or 'streaming'
        pagesize : int, optional
            The maximum number of events in a page
        filter : string, optional
            Functional filter to subset events
        sort : string, optional
            Sort order for the events (updating mode only)
        format : string, optional
            The format of the received data: 'xml', 'json', 'csv', 'properties'
        separator : string, optional
            The separator to use between events in the 'properties' format
        interval : int, optional
            Interval between event sends in milliseconds
        schema : bool, optional
            Should the schema be sent with the first event?
        on_event : callable, optional
            The object to call for events. The argument to this object will
            be a DataFrame of the events that occurred.
        on_message : callable, optional
            The object to call for each websocket message
        on_error : callable, optional
            The object to call for each websocket error
        on_close : callable, optional
            The object to call when the websocket is opened
        on_open : callable, optional
            The object to call when the websocket is closed

        Examples
        --------
        Define a callback function for events

        >>> def on_event(event):
        ...     print(event.columns)

        Create a window subscrber

        >>> sub = win.create_subscriber(on_event=on_event)

        Start processing events (runs a background thread)

        >>> sub.start()

        Stop processing events (stops background thread)

        >>> sub.stop()

        Returns
        -------
        :class:`Subscriber`

        '''
        return Subscriber(self, mode=mode, pagesize=pagesize, filter=filter,
                          sort=sort, format=format, separator=separator,
                          interval=interval, schema=schema,
                          on_event=on_event, on_message=on_message,
                          on_error=on_error, on_close=on_close, on_open=on_open)

    def create_publisher(self, blocksize=1, rate=0, pause=0,
                         dateformat='%Y%m%dT%H:%M:%S.%f', opcode='insert',
                         format='csv', separator=None):
        '''
        Create a publisher for the given window

        Parameters
        ----------
        blocksize : int, optional
            Number of events to put into an event block
        rate : int, optional
            Maximum number of events to inject per second
        pause : int, optional
            Number of milliseconds to pause between each injection of events
        dateformat : string, optional
            Format for date fields
        format : string, optional
            The data format of inputs: 'csv', 'xml', 'json', 'properties'
        separator : string, optional
            The separator to use between events in the 'properties' format
        opcode : string, optional
            Opcode to use if an input event does not include one:
            'insert', 'upsert', 'delete'

        Examples
        --------
        Create a publisher instance

        >>> pub = win.create_publisher(pause=200)

        Publish CSV data to subscriber

        >>> pub.send('1,2,3')

        Close the connection

        >>> pub.close()

        Returns
        -------
        :class:`Publisher`

        '''
        return Publisher(self, blocksize=blocksize, rate=rate, pause=pause,
                         dateformat=dateformat, format=format,
                         separator=separator)

    def publish_events(self, data, blocksize=1, rate=0, pause=0,
                       dateformat='%Y%m%dT%H:%M:%S.%f', opcode='insert',
                       format=None, separator=None):
        '''
        Publish events to the window

        Parameters
        ----------
        data : string or file-like or DataFrame
            The event data to publish
        blocksize : int, optional
            Number of events to put into an event block
        rate : int, optional
            Maximum number of events to inject per second
        pause : int, optional
            Number of milliseconds to pause between each injection of events
        dateformat : string, optional
            Format for date fields
        opcode : string, optional
            Opcode to use if an input event does not include one:
            'insert', 'upsert', 'delete'
        format : string, optional
            The data format of inputs: 'csv', 'xml', 'json', 'properties'
        separator : string
            The separator string to use between events in 'properties' format

        Examples
        --------
        Publish CSV data to window with at one event per 200 milliseconds.

        >>> with open('data.csv') as csvfile:
        ...     win.publish_events(csvfile, pause=200)

        Publish DataFrame data to window, only columns that match the
        window's schema will be pubished.

        >>> win.publish_events(dframe, pause=200)

        '''
        data_file = None

        if isinstance(data, pd.DataFrame):
            index = False
            if hasattr(self, 'schema') and self.schema.fields:
                data = data.reset_index().loc[:, list(self.schema.fields.keys())]
            else:
                if len(data.index.names) > 1 or data.index.names[0] is not None:
                    index = True
            data = data.to_csv(header=False, index=index, na_rep='')
            format = 'csv'

        elif isinstance(data, six.string_types):
            if os.path.isfile(data):
                data_file = open(data, 'r')
                data = data_file.read()

        elif hasattr(data, 'read'):
            data = data.read()

        else:
            raise TypeError('Unknown data type: %s' % data)

        if format is None:
            if data.startswith('<'):
                format = 'xml'
            elif data.startswith('{'):
                format = 'json'
            elif re.match(r'\w+=', data):
                format = 'properties'
            else:
                format = 'csv'

        try:
            pub = self.create_publisher(blocksize=blocksize, rate=rate, pause=pause,
                                        dateformat=dateformat, opcode=opcode,
                                        format=format, separator=separator)
            pub.send(data)
            pub.close()

        finally:
            if data_file is not None:
                data_file.close()

    def _get_event_horizon(self, value):
        '''
        Set a timespan or deadline for event collection

        Parameters
        ----------
        value : datetime or date or time or timedelta or int
            The deadline or timespan for event collection.  If a datetime,
            date, or time is given, streaming stops at the next event that
            occurs after that value.  If a timedelta is given, streaming stops
            at the first event after the timespan given in the timedelta
            passes.  An integer value, specifies a maximum number of events
            to retrieve.

        '''
        if value is None:
            return

        if not isinstance(value, (list, tuple)):
            value = [value]

        out = []
        for item in value:
            if item is None:
                pass
            elif isinstance(item, (datetime.datetime, six.string_types)):
                out.append(item)
            elif isinstance(item, datetime.timedelta):
                out.append(datetime.datetime.now() + item)
            elif isinstance(item, datetime.date):
                out.append(datetime.datetime(item.year, item.month, item.day))
            elif isinstance(item, datetime.time):
                today = datetime.datetime.combine(datetime.date.today(), datetime.time())
                tomorrow = datetime.datetime.combine(datetime.date.today() +
                                                     datetime.timedelta(days=1), datetime.time())
                if today > datetime.datetime.now():
                    out.append(tomorrow)
                else:
                    out.append(today)
            elif isinstance(item, (int, float)):
                out.append(int(item))
            else:
                raise TypeError('Unknown type for event horizon: %s' % value)

        return out

#   def _get_sort_order(self, value):
#      if value:
#          return value
#      if self.schema.fields:
#           for field in self.schema.fields.values():
#               if field.key:
#                   return '%s:descending' % field.name

    def subscribe(self, mode='streaming', pagesize=50, filter=None,
                  sort=None, interval=None, limit=None, horizon=None, reset=True):
        '''
        Subscribe to events

        When invoked, this method creates a :class:`Subscriber` object to collect
        events for the window.  The data in these events are stored in
        the :attr:`data` attribute.

        To stop event processing, use the :meth:`unsubscribe` method.

        Parameters
        ----------
        mode : string, optional
            The mode of subscriber: 'updating' or 'streaming'
        pagesize : int, optional
            The maximum number of events in a page
        filter : string, optional
            Functional filter to subset events
        sort : string, optional
            Sort order for the events (updating mode only)
        interval : int, optional
            Interval between event sends in milliseconds
        limit : int, optional
            The maximum number of rows of data to keep in the internal
            DataFrame object.
        horizon : int or datetime.datetime or string, optional
            Specifies a condition that stops the subscriber.
            If an int, the subscriber stops after than many events.
            If a datetime.datetime, the subscriber stops after the specified
            date and time.  If a string, the string is an expression
            applied to the event using the :meth:`DataFrame.query`
            method.  If that query returns any number of rows, the
            subscriber is stopped.
        reset : bool, optional
            If True, the internal data is reset on subsequent calls
            to the :meth:`subscribe` method.

        See Also
        --------
        :meth:`unsubscribe`
        :class:`Subscriber`

        '''
        self.unsubscribe()

        if self.data is None or reset:
            self.data = get_dataframe(self)

        state = dict(total=0, limit=limit or sys.maxsize,
                     horizon=self._get_event_horizon(horizon))

        if self.data is None:
            self.data = get_dataframe(self)

        def on_event(ws, event):
            if state['horizon'] is not None:
                for item in state['horizon']:
                    if isinstance(item, int):
                        if state['total'] >= item:
                            self.unsubscribe()
                            return
                    if isinstance(item, datetime.datetime):
                        if item < datetime.datetime.now():
                            self.unsubscribe()
                            return
                    if isinstance(item, six.string_types):
                        if len(event.query(item)):
                            self.unsubscribe()
                            return
            for method, args, kwargs in self.event_transformers:
                if isinstance(method, six.string_types):
                    event = getattr(event, method)(*args, **kwargs)
                else:
                    args = [event] + list(args)
                    event = method(*args, **kwargs)
            columns = self.data.columns
            self.data = pd.concat([self.data, event],
                                  **CONCAT_OPTIONS)[-state['limit']:][columns]
            state['total'] += len(event)

        self._subscriber = self.create_subscriber(mode=mode, pagesize=pagesize,
                                                  filter=filter, sort=sort,
                                                  interval=interval, format='xml',
                                                  on_event=on_event)
        self._subscriber.start()

    def unsubscribe(self):
        '''
        Stop event processing

        See Also
        --------
        :meth:`subscribe`

        '''
        if self._subscriber is not None:
            self._subscriber.close()
        self._subscriber = None

    def apply_transformers(self, data):
        '''
        Apply current set of data transformers to given data

        Parameters
        ----------
        data : DataFrame
            The data set to apply transformers to

        Examples
        --------
        >>> win.add_event_transformer('clip', upper=10, lower=0)

        >>> dframe = pd.DataFrame([[-1, 2, 3], [10, 20, 30]])
        >>> dframe
            0   1   2
        0  -1   2   3
        1  10  20  30

        >>> win.apply_transformers(dframe)
            0   1   2
        0   0   2   3
        1  10  10  10

        See Also
        --------
        :meth:`add_event_transformer`

        Returns
        -------
        :class:`pandas.DataFrame`

        '''
        for method, args, kwargs in self.event_transformers:
            if isinstance(method, six.string_types):
                data = getattr(data, method)(*args, **kwargs)
            else:
                args = [data] + list(args)
                data = method(*args, **kwargs)
        return data

    @property
    def contquery(self):
        '''
        The name of the continuous query the window is associated with

        Returns
        -------
        string

        '''
        return self._contquery

    @contquery.setter
    def contquery(self, value):
        '''
        Set the continous_query value

        Parameters
        ----------
        value : string or ContinousQuery
            The name of the continuous query.  If a :class:`ContinousQuery`
            object is used, the name will be extracted from it.

        '''
        self._contquery = getattr(value, 'name', value)
        if hasattr(value, 'project'):
            self.project = value.project

    @property
    def project(self):
        '''
        The name of the project the window is associated with

        Returns
        -------
        string

        '''
        return self._project

    @project.setter
    def project(self, value):
        '''
        Set the project value

        Parameters
        ----------
        value : string or Project
            The name of the project.  If a :class:`Project`
            object is used, the name will be extracted from it.

        '''
        self._project = getattr(value, 'name', value)

    def create_event_generator(self, data=None, name=None, overwrite=False):
        '''
        Create an event generator

        Parameters
        ----------
        data : CSV or URL or DataFrame, optional
            The events to inject
        name : string, optional
            The name of the event generator.  If no name is given,
            a name will be generated.
        overwrite : bool, optional
            Should an existing event generator with the same name
            be overwritten?

        Examples
        --------
        Create the event generator

        >>> egen = win.create_event_generator('file:///path/to/file.csv')

        Insert events at a rate of one every 200 milliseconds

        >>> egen.start(pause=200)

        See Also
        --------
        :class:`EventGenerator`

        Returns
        -------
        :class:`EventGenerator`

        '''
        from ..evtgen import EventGenerator
        gen = EventGenerator(name=name)
        gen.session = self.session
        gen.publish_target = self
        if data is not None:
            gen.event_data = data
            gen.save(overwrite=overwrite)
        return gen

    def add_targets(self, *windows, **kwargs):
        '''
        Add windows as targets

        Parameters
        ----------
        windows : one-or-more Windows or window names
            The windows to use as targets
        role : string, optional
            The role of the connection
        slot : string, optional
            Indicates the slot number to use from the splitting
            function for the window.

        Examples
        --------
        >>> train_w.add_target(score_w, role='model')
        >>> win.add_target(train_w, role='data').add_target(score_w, role='data')

        See Also
        --------
        :meth:`delete_target`
        :meth:`delete_targets`

        Returns
        -------
        ``self``

        '''
        self.delete_targets(*windows)
        for item in windows:
            self.targets.add(Target(name=getattr(item, 'name', item).split('.')[-1],
                             role=kwargs.get('role', None), slot=kwargs.get('slot', None)))

    add_target = add_targets

    def delete_targets(self, *windows):
        '''
        Remove windows as targets

        Parameters
        ----------
        windows : one-or-more Windows or window names
            The windows to remove as targets

        Examples
        --------
        >>> win.delete_targets(train_w, score_w)

        See Also
        --------
        :meth:`add_target`
        :meth:`add_targets`

        Returns
        -------
        ``self``

        '''
        windows = set([getattr(x, 'name', x).split('.')[-1]
                       for x in windows])
        for item in windows:
            for target in set(self.targets):
                if target.name in windows:
                    self.targets.remove(target)

    delete_target = delete_targets

    @property
    def fullname(self):
        '''
        The fully-qualified name of the window (e.g., project.contquery.window)

        Returns
        -------
        string

        '''
        return '%s.%s.%s' % (self.project, self.contquery, self.name)

    @property
    def url(self):
        '''
        URL of the window

        Returns
        -------
        string

        '''
        self._verify_project()
        return urllib.parse.urljoin(self.base_url, '%s/%s/%s/' %
                                    (self.project, self.contquery, self.name))

    def _server_info(self):
        ''' Retrieve the server information '''
        return get_server_info(self)

    def _verify_project(self):
        '''
        Verify that the project and continuous query attributes are set

        '''
        if not self.contquery:
            raise ValueError('This window is not associated with a continuous query.')
        if not self.project:
            raise ValueError('This window is not associated with a project.')

    def copy(self, deep=False):
        '''
        Return a copy of the object

        Parameters
        ----------
        deep : bool, optional
            Should sub-objects be copied as well?

        Examples
        --------
        Shallow copy

        >>> new_win = win.copy()

        Deep copy

        >>> new_win = win.copy(deep=True)

        Returns
        -------
        :class:`Window`

        '''
        out = type(self)()

        out.session = self.session
        out.contquery = self.contquery
        out.project = self.project
        out.targets = set(self.targets)
        out.description = self.description

        for key, value in self._get_attributes(use_xml_values=False).items():
            setattr(out, key, value)

        if self.data is not None:
            out.data = self.data.copy(deep=True)

        out.event_transformers = list(self.event_transformers)

        features = WindowFeature.__subclasses__()
        for item in inspect.getmro(type(self)):
            if item in features:
                item._copy_feature(self, out, deep=deep)

        return out

    def __copy__(self):
        return self.copy(deep=False)

    def __deepcopy__(self, memo):
        return self.copy(deep=True)

    @classmethod
    def from_xml(cls, data, contquery=None, project=None, session=None):
        '''
        Construct window from XML definition

        Parameters
        ----------
        data : xml-string or ElementTree.Element
            XML window definition
        contquery : string or ContinuousQuery, optional
            The continuous query to associate the window with
        project : string or Project, optional
            The project to associate the window with
        session : requests.Session, optional
            The session object

        Examples
        --------
        Create a window from XML

        >>> win = Window.from_xml('<window-source name="w_data" insert-only="true">'
        ...                       '  <schema>'
        ...                       '    <fields>'
        ...                       '      <field name="id" type="int64" key="true" />'
        ...                       '      <field name="time" type="double" />'
        ...                       '    </fields>'
        ...                       '  </schema>'
        ...                       '</window-source>')
        >>> win
        SourceWindow(name='w_data', contquery=None, project=None)

        Display the schema

        >>> win.schema
        id*:int64,time:double

        Returns
        -------
        :class:`Window` subclass

        '''
        if isinstance(data, six.string_types):
            data = xml.from_xml(data)

        if cls in [BaseWindow, Window]:
            cls = cls.window_classes[data.tag]
            return cls.from_xml(data, contquery=contquery,
                                project=project, session=session)

        out = cls()
        out.session = session
        out.project = getattr(project, 'name', project)
        out.contquery = getattr(contquery, 'name', contquery)

        out._set_attributes(data.attrib)

        for item in data.findall('./description'):
            out.description = item.text

        for item in data.findall('./schema'):
            out.schema = Schema.from_xml(item, session=session)
        for item in data.findall('./schema-string'):
            out.schema = Schema.from_schema_string(item, session=session)

        for item in list(inspect.getmro(cls))[1:]:
            if issubclass(item, WindowFeature):
                item._feature_from_element(out, data)

        return out

    from_element = from_xml

    def to_element(self, query=None):
        '''
        Export window definition to ElementTree.Element

        Parameters
        ----------
        query : ContinuousQuery, optional
            The parent query object of the window

        Returns
        -------
        :class:`ElementTree.Element`

        '''
        attrs = xml.get_attrs(self, exclude='project')
        attrs.pop('contquery', None)
        out = xml.new_elem('window-%s' % self.window_type, attrs)

        if self.description:
            xml.add_elem(out, 'description', text_content=self.description)

        features = WindowFeature.__subclasses__()
        for item in inspect.getmro(type(self)):
            if item in features:
                elem = item._feature_to_element(self)
                if elem is None:
                    pass
                elif isinstance(elem, list):
                    for subelem in elem:
                        xml.add_elem(out, subelem)
                else:
                    xml.add_elem(out, elem)

        connectors_to_end(out)

        return out

    def to_xml(self, pretty=False, query=None):
        '''
        Export window definition to XML

        Parameters
        ----------
        pretty : bool, optional
            Should the output include whitespace for readability?
        query : ContinuousQuery, optional
            The parent query object of this window

        Examples
        --------
        >>> win.to_xml(pretty=True)
        <window-source insert-only="true" name="w_data">
          <schema>
            <fields>
              <field key="true" name="id" type="int64" />
              <field key="false" name="time" type="double" />
            </fields>
          </schema>
        </window-source>

        Returns
        -------
        string

        '''
        return xml.to_xml(self.to_element(query=query), pretty=pretty)

    def save_xml(self, dest, mode='w', pretty=True, **kwargs):
        '''
        Save the window XML to a file

        Parameters
        ----------
        dest : string or file-like
            The destination of the XML content
        mode : string, optional
            The write mode for the output file (only used if `dest` is a string)
        pretty : boolean, optional
            Should the XML include whitespace for readability?

        '''
        if isinstance(dest, six.string_types):
            with open(dest, mode=mode, **kwargs) as output:
                output.write(self.to_xml(pretty=pretty))
        else:
            dest.write(self.to_xml(pretty=pretty))

    def to_graph(self, graph=None, schema=True):
        '''
        Export window definition to graphviz.Digraph

        Parameters
        ----------
        graph : graphviz.Graph, optional
            The parent graph to add to
        schema : bool, optional
            Should window schemas be included?

        Returns
        -------
        :class:`graphviz.Digraph`

        '''
        try:
            import graphviz as gv
        except ImportError:
            raise ImportError('The graphviz module is required for exporting to graphs.')

        if graph is None:
            graph = gv.Digraph(format='svg')
            graph.attr('node', shape='rect')
            graph.attr('graph', rankdir='LR')

        if schema and hasattr(self, 'schema'):
            key_fmt = '<tr><td>&#x1f511;<font color="#c49810">%s</font></td></tr>'
            col_fmt = '<tr><td><font color="#808080">%s</font></td></tr>'
            out = self.schema.to_element()
            fields = [x.attrib for x in out.findall('.//field')]
            label = []
            label.append('<<table border="0" cellspacing="0" cellpadding="0">')
            label.append('<tr><td>%s</td></tr>' % self.name)
            label.append('<tr><td> </td></tr>')
            n_fields = 0
            for field in fields:
                fmt = (field.get('key', 'false') == 'true') and key_fmt or col_fmt
                if get_option('display.show_field_type'):
                    label.append(fmt % ' : '.join((field['name'], field['type'])))
                else:
                    label.append(fmt % field['name'])
                n_fields += 1
                if n_fields >= get_option('display.max_fields'):
                    if len(fields) - n_fields > 1:
                        label.append(('<tr><td><font color="#808080">[%s more]'
                                      '</font></td></tr>') % (len(fields) - n_fields))
                        break
            label.append('</table>>')
            label = ''.join(label)
        else:
            label = self.name

        graph.node(self.fullname, label,
                   style='filled,bold', color='#58a0d3',
                   fillcolor='#c8f0ff', fontcolor='black',
                   margin='.25,.17', fontname='helvetica')

        return graph

    def _repr_svg_(self):
        try:
            return scale_svg(self.to_graph()._repr_svg_())
        except ImportError:
            raise AttributeError('_repr_svg_')

    def __getitem__(self, key):
        return self.data[key]

    def __getattr__(self, name):
        if name in self._schema.fields or name in dir(pd.DataFrame):
            return getattr(self.data, name)
        raise AttributeError(name)

    def __setitem__(self, name, value):
        raise RuntimeError('Events are read-only')

    def __delitem__(self, name):
        raise RuntimeError('Events are read-only')

    def __len__(self):
        if self.data is None:
            return 0
        return len(self.data)

    def __iter__(self):
        if self.data is None:
            return iter([])
        return iter(self.data)

    def enable_tracing(self):
        ''' Enable console tracing for the window '''
        self._verify_project()
        self._put(urllib.parse.urljoin(self.base_url,
                                       'windows/%s/%s/%s/state' % (self.project,
                                                                   self.contquery,
                                                                   self.name)),
                  params=get_params(value='tracingOn'))

    def disable_tracing(self):
        ''' Disable console tracing for the window '''
        self._verify_project()
        self._put(urllib.parse.urljoin(self.base_url,
                                       'windows/%s/%s/%s/state' % (self.project,
                                                                   self.contquery,
                                                                   self.name)),
                  params=get_params(value='tracingOff'))

    def get_events(self, filter=None, sort_by=None, limit=None):
        '''
        Retrieve events from the window

        Parameters
        ----------
        filter : string or list-of-strings, optional
            Functional filter indicating the events to return
        sort_by : string, optional
            Field to sort results on.  It should be in the form
            ``field:order``, where ``field`` is the field name and
            ``order`` is either ``ascending`` or ``descending``.
        limit : int, optional
            Maximum number of events to return

        See Also
        --------
        :meth:`subscribe`
        :meth:`create_subscriber`

        Returns
        -------
        :class:`pandas.DataFrame`

        '''
        self._verify_project()
        out = get_events(self, self._get(urllib.parse.urljoin(self.base_url,
                                         'events/%s/%s/%s/' % (self.project,
                                                               self.contquery,
                                                               self.name)),
                                         params=get_params(filter=filter,
                                                           sort_by=sort_by,
                                                           limit=limit)),
                         server_info=get_server_info(self))
        if out:
            return list(out.values())[0]

        out = pd.DataFrame(columns=self.schema.fields.keys())
        index = [x.name for x in self.schema.fields.values() if x.key]
        if index:
            out = out.set_index(index)
        return out

    def get_pattern_events(self, sort_by=None, limit=None):
        '''
        Retrieve events residing in open patterns in the window

        Parameters
        ----------
        sort_by : string, optional
            Field to sort results on.  It should be in the form
            ``field:order``, where ``field`` is the field name and
            ``order`` is either ``ascending`` or ``descending``.
        limit : int, optional
            Maximum number of events to return

        Returns
        -------
        :class:`pandas.DataFrame`

        '''
        self._verify_project()
        out = get_events(self,
                         self._get(urllib.parse.urljoin(self.base_url,
                                                        'patternEvents/%s/%s/%s/' %
                                                        (self.project,
                                                         self.contquery,
                                                         self.name)),
                                   params=get_params(sort_by=sort_by, limit=limit)),
                         server_info=get_server_info(self))
        if out:
            return list(out.values())[0]

        out = pd.DataFrame(columns=self.schema.fields.keys())
        index = [x.name for x in self.schema.fields.values() if x.key]
        if index:
            out = out.set_index(index)
        return out

    def __str__(self):
        if self.data is None:
            return '%s(name=%s, contquery=%s, project=%s)' % \
                   (type(self).__name__, repr(self.name),
                    repr(self.contquery), repr(self.project))
        return str(self.data)

    def __repr__(self):
        return str(self)

    def to_data_callback(self, x=None, y=None, extra=None, max_data=None,
                         interval=None, var_generator=None):
        '''
        Create a data callback function for streaming plots

        Parameters
        ----------
        x : string, optional
            x-axis variable name
        y : string or list-of-strings, optional
            y-axis variable names
        extra : string or list-of-strings, optional
            Extra variable names
        max_data : int, optional
            The maximum number of data points to return
        interval : int, optional
            The subscriber interval
        var_generator : callable, optional
            Function to call on new data to generate additional variables

        Examples
        --------
        Define function to generate colors

        >>> def var_gen(data):
        ...     return dict(x_colors=(data['x'] > 5).map({True: 'green',
        ...                                               False: 'red'}))

        Create callback function

        >>> func = win.to_data_callback(x='time', y=['x', 'y', 'z'], max_data=5,
        ...                             var_generator=var_gen)
        >>> func()
        {'time': [11.46, 11.49, 11.52, 11.55, 11.58],
         'x': [-2.9829, -2.9829, -3.0237, -2.9556, -2.9148],
         'x_colors': id
         375    red
         376    red
         377    red
         378    red
         379    red
         Name: x, dtype: object,
         'y': [9.112, 9.1529, 9.2346, 9.1937, 9.1529],
         'z': [-1.1169, -1.1169, -1.076, -1.0351, -1.1169]}

        Returns
        -------
        function

        '''
        if var_generator is None:
            var_generator = lambda x: dict()
        elif isinstance(var_generator, dict):
            var_generator = functools.partial(var_mapper, mapping=var_generator)

        empty = self.apply_transformers(get_dataframe(self))
        state = dict(events=empty, reset=False)

        def on_event(sock, event):
            try:
                if state['reset']:
                    state['reset'] = False
                    state['events'] = self.apply_transformers(event)[:max_data]
                else:
                    state['events'] = pd.concat([state['events'],
                                                 self.apply_transformers(event)],
                                                **CONCAT_OPTIONS)[:max_data]
            except:
                import traceback
                traceback.print_exc()

        subscriber = self.create_subscriber(mode='streaming',
                                            format='xml',
                                            # interval=interval,
                                            sort='%s:descending' % x,
                                            on_event=on_event)
        subscriber.start()

        if y is None:
            y = []
        elif isinstance(y, six.string_types):
            y = [y]

        if x is None:
            x = []
        elif isinstance(x, six.string_types):
            x = [x]

        if extra is None:
            extra = []
        elif isinstance(extra, six.string_types):
            extra = [extra]

        def data_callback(initial=False, max_data=None, terminate=False):
            if terminate:
                subscriber.stop()
                state['events'] = None
                return

            data = state['events']
            state['reset'] = True

            if data is None:
                return {}

            out = {}
            for name in x + y + extra:
                if name in data.index.names:
                    out[name] = data.index.get_level_values(name).tolist()
                else:
                    out[name] = data[name].tolist()

            out.update(var_generator(data))

            return out

        return data_callback

    def _get_colors(self, num, palette=None):
        '''
        Generate colors for plot

        Parameters
        ----------
        palette : function
            Function that takes an integer and returns that number of colors
        num : int
            The number of colors to return

        Returns
        -------
        list-of-strings

        '''
        from bokeh.palettes import viridis
        palette = palette or viridis
        return palette(num + 1)

    def streaming_line(self, x, y, steps=1e5, interval=1000,
                       max_data=100, var_generator=None, **kwargs):
        '''
        Display a streaming line plot in a Jupyter notebook

        Notes
        -----
        Streaming figures are only supported in Jupyter notebooks.

        Parameters
        ----------
        x : string
            The x-axis variable name
        y : string or list-of-strings
            The y-axis variable names
        steps : int, optional
            The maximum number of steps to iterate
        interval : int, optional
            The length of each step in milliseconds
        max_data : int, optional
            The maximum number of data points to display
        var_generator : callable, optional
            Function to call on new data to generate additional variables
        **kwargs : keyword arguments, optional
            Arguments to send to the :func:`bokeh.plotting.figure` function

        Examples
        --------
        Create the streaming figure.  Evaluating this in a Jupyter notebook
        will display the streaming plot.

        >>> fig = dataw.streaming_line(x='time', y=['x', 'y', 'z'])
        >>> fig
        <esppy.plotting.StreamingChart at 0x7f98c0250e10>

        See Also
        --------
        :meth:`streaming_scatter`

        Returns
        -------
        :class:`StreamingChart`

        '''
        y = listify(y)

        chart_params, mthd_params = split_chart_params(**kwargs)

        out = StreamingChart(self, var_generator=var_generator,
                             steps=steps, interval=interval,
                             max_data=max_data, **chart_params)

        for item, params in zip(y, param_iter(mthd_params)):
            out.line(x, item, **params)

        return out

    def streaming_area(self, x, y, steps=1e5, interval=1000,
                       max_data=100, var_generator=None, **kwargs):
        '''
        Display a streaming area plot in a Jupyter notebook

        Notes
        -----
        Streaming figures are only supported in Jupyter notebooks.

        Parameters
        ----------
        x : string
            The x-axis variable name
        y : string or list-of-strings
            The y-axis variable names
        steps : int, optional
            The maximum number of steps to iterate
        interval : int, optional
            The length of each step in milliseconds
        max_data : int, optional
            The maximum number of data points to display
        var_generator : callable, optional
            Function to call on new data to generate additional variables
        **kwargs : keyword arguments, optional
            Arguments to send to the :func:`bokeh.plotting.figure` function

        Examples
        --------
        Create the streaming figure.  Evaluating this in a Jupyter notebook
        will display the streaming plot.

        >>> fig = dataw.streaming_area(x='time', y=['x', 'y', 'z'])
        >>> fig
        <esppy.plotting.StreamingChart at 0x7f98c0250e10>

        See Also
        --------
        :meth:`streaming_scatter`

        Returns
        -------
        :class:`StreamingChart`

        '''
        y = listify(y)

        chart_params, mthd_params = split_chart_params(**kwargs)

        out = StreamingChart(self, var_generator=var_generator,
                             steps=steps, interval=interval,
                             max_data=max_data, **chart_params)

        for item, params in zip(y, param_iter(mthd_params)):
            out.area(x, item, **params)

        return out

    def streaming_hist(self, centers=None, heights=None, steps=1e5, interval=1000,
                       var_generator=None, **kwargs):
        '''
        Display a streaming histogram in a Jupyter notebook

        Notes
        -----
        Streaming figures are only supported in Jupyter notebooks.

        Parameters
        ----------
        centers : list-of-strings or regex, optional
            The names of the columns containing the bin centers
        heights : list-of-strings or regex, optional
            The names of the columns containing the bin heights
        steps : int, optional
            The maximum number of steps to iterate
        interval : int, optional
            The length of each step in milliseconds
        max_data : int, optional
            The maximum number of data points to display
        var_generator : callable, optional
            Function to call on new data to generate additional variables
        **kwargs : keyword arguments, optional
            Arguments to send to the :func:`bokeh.plotting.figure` function

        See Also
        --------
        :meth:`streaming_area`

        Returns
        -------
        :class:`StreamingChart`

        '''
        if centers is None:
            centers = getattr(self, 'output_map', {}).get('binCentersOut', None)
        if heights is None:
            heights = getattr(self, 'output_map', {}).get('binHeightsOut', None)

        max_data = getattr(self, 'parameters', {}).get('nBins', 0)

        use_arrays = False
        centers_start = 0
        centers_end = 0
        heights_start = 0
        heights_end = 0

        # Check for array types centers[#-#]
        def extract_range(value):
            ''' Extract name and number range '''
            parts = re.match(r'^\s*(\w+)\s*\[\s*(\d+)\s*-\s*(\d+)\s*\]\s*$', value)
            name = parts.group(1)
            start = int(parts.group(2))
            end = int(parts.group(3))
            return name, start, end

        if isinstance(centers, six.string_types) and \
                re.match(r'^\s*\w+\s*\[\s*\d+\s*-\s*\d+\s*\]\s*$', centers):
            centers, centers_start, centers_end = extract_range(centers)
            use_arrays = True

        if isinstance(heights, six.string_types) and \
                re.match(r'^\s*\w+\s*\[\s*\d+\s*-\s*\d+\s*\]\s*$', heights):
            heights, heights_start, heights_end = extract_range(heights)
            use_arrays = True

        if not centers:
            raise ValueError('No center column names were specified')
        if not heights:
            raise ValueError('No height column names were specified')

        # Centers / heights are an array in one cell
        if use_arrays:
            if (heights_end - heights_start) != (centers_end - centers_start):
                raise ValueError('Center and height arrays are different lengths')

            if max_data == 0:
                max_data = heights_start - heights_end + 1

        # Centers / heights are individual columns
        else:
            centers = list(centers)
            heights = list(heights)

            if len(centers) != len(heights):
                raise ValueError('Center and height arrays are different lengths')

            if max_data == 0:
                max_data = len(centers)

        def hist_var_generator(data, centers=centers, heights=heights,
                               var_generator=var_generator):
            if not len(data):
                out = {'_centers_': [], '_heights_': []}
            else:
                out = {'_centers_': list(data[centers].iloc[-1]),
                       '_heights_': list(data[heights].iloc[-1])}
            if var_generator is not None:
                out.update(var_generator(data))
            return out

        kwargs.setdefault('x_axis_type', 'linear')
        kwargs.setdefault('line_width', 2)
        kwargs.setdefault('point_radius', 3)
        kwargs.setdefault('label', 'height')

        return self.streaming_area(x='_centers_', y='_heights_',
                                   interval=interval, steps=steps,
                                   var_generator=hist_var_generator,
                                   max_data=max_data, **kwargs)

    def streaming_scatter(self, x, y, steps=1e5, interval=1000,
                          max_data=100, var_generator=None, **kwargs):
        '''
        Display a streaming scatter plot in a Jupyter notebook

        Notes
        -----
        Streaming figures are only supported in Jupyter notebooks.

        Parameters
        ----------
        x : string
            The x-axis variable name
        y : string or list-of-strings
            The y-axis variable names
        steps : int, optional
            The maximum number of steps to iterate
        interval : int, optional
            The length of each step in milliseconds
        max_data : int, optional
            The maximum number of data points to display
        var_generator : callable, optional
            Function to call on new data to generate additional variables
        **kwargs : keyword arguments, optional
            Arguments to send to the :func:`bokeh.plotting.figure` function

        Examples
        --------
        Create the streaming figure.  Evaluating this in a Jupyter notebook
        will display the streaming plot.

        >>> fig = dataw.streaming_scatter(x='time', y=['x', 'y', 'z'])
        >>> fig
        <esppy.plotting.StreamingChart at 0x7f98c0250e10>

        See Also
        --------
        :meth:`streaming_line`

        Returns
        -------
        :class:`StreamingChart`

        '''
        y = listify(y)

        chart_params, mthd_params = split_chart_params(**kwargs)

        out = StreamingChart(self, var_generator=var_generator,
                             steps=steps, interval=interval,
                             max_data=max_data, **chart_params)

        for item, params in zip(y, param_iter(mthd_params)):
            out.scatter(x, item, **params)

        return out

    def streaming_bubble(self, x, y, radius, steps=1e5, interval=1000,
                         max_data=100, var_generator=None, **kwargs):
        '''
        Display a streaming bubble plot in a Jupyter notebook

        Notes
        -----
        Streaming figures are only supported in Jupyter notebooks.

        Parameters
        ----------
        x : string
            The x-axis variable name
        y : string or list-of-strings
            The y-axis variable names
        r : string or list-of-strings
            The variable name that gives the point radius
        steps : int, optional
            The maximum number of steps to iterate
        interval : int, optional
            The length of each step in milliseconds
        max_data : int, optional
            The maximum number of data points to display
        var_generator : callable, optional
            Function to call on new data to generate additional variables
        **kwargs : keyword arguments, optional
            Arguments to send to the :func:`bokeh.plotting.figure` function

        Examples
        --------
        Create the streaming figure.  Evaluating this in a Jupyter notebook
        will display the streaming plot.

        >>> fig = dataw.streaming_scatter(x='time', y=['x', 'y', 'z'])
        >>> fig
        <esppy.plotting.StreamingChart at 0x7f98c0250e10>

        See Also
        --------
        :meth:`streaming_line`

        Returns
        -------
        :class:`StreamingChart`

        '''
        y = listify(y)
        radius = listify(radius)

        if len(y) > 1 and len(radius) == 1:
            radius = radius * len(y)

        chart_params, mthd_params = split_chart_params(**kwargs)

        out = StreamingChart(self, var_generator=var_generator,
                             steps=steps, interval=interval,
                             max_data=max_data, **chart_params)

        for item, rad, params in zip(y, radius, param_iter(mthd_params)):
            out.bubble(x, item, rad, **params)

        return out

    def streaming_bar(self, x, y, steps=1e5, interval=1000,
                      max_data=100, var_generator=None, displayed_labels=None, **kwargs):
        '''
        Display a streaming bar plot in a Jupyter notebook

        Notes
        -----
        Streaming figures are only supported in Jupyter notebooks.

        Parameters
        ----------
        x : string
            The x-axis variable name
        y : string or list-of-strings
            The y-axis variable names
        steps : int, optional
            The maximum number of steps to iterate
        interval : int, optional
            The length of each step in milliseconds
        max_data : int, optional
            The maximum number of data points to display
        var_generator : callable, optional
            Function to call on new data to generate additional variables
        **kwargs : keyword arguments, optional
            Arguments to send to the :func:`bokeh.plotting.figure` function

        Examples
        --------
        Create the streaming figure.  Evaluating this in a Jupyter notebook
        will display the streaming plot.

        >>> fig = dataw.streaming_bar(x='time', y=['x', 'y', 'z'])
        >>> fig
        <esppy.plotting.StreamingChart at 0x7f98c0250e10>

        See Also
        --------
        :meth:`streaming_line`

        Returns
        -------
        :class:`StreamingChart`

        '''
        y = listify(y)

        chart_params, mthd_params = split_chart_params(**kwargs)

        out = StreamingChart(self, var_generator=var_generator,
                             steps=steps, interval=interval,
                             max_data=max_data, **chart_params)

        for item, params in zip(y, param_iter(mthd_params)):
            out.bar(x, item, displayed_labels=displayed_labels, **params)

        return out

    def streaming_hbar(self, x, y, steps=1e5, interval=1000,
                       max_data=100, var_generator=None, **kwargs):
        '''
        Display a streaming horizontal bar plot in a Jupyter notebook

        Notes
        -----
        Streaming figures are only supported in Jupyter notebooks.

        Parameters
        ----------
        x : string
            The x-axis variable name
        y : string or list-of-strings
            The y-axis variable names
        steps : int, optional
            The maximum number of steps to iterate
        interval : int, optional
            The length of each step in milliseconds
        max_data : int, optional
            The maximum number of data points to display
        var_generator : callable, optional
            Function to call on new data to generate additional variables
        **kwargs : keyword arguments, optional
            Arguments to send to the :func:`bokeh.plotting.figure` function

        Examples
        --------
        Create the streaming figure.  Evaluating this in a Jupyter notebook
        will display the streaming plot.

        >>> fig = dataw.streaming_bar(x='time', y=['x', 'y', 'z'])
        >>> fig
        <esppy.plotting.StreamingChart at 0x7f98c0250e10>

        See Also
        --------
        :meth:`streaming_line`

        Returns
        -------
        :class:`StreamingChart`

        '''
        y = listify(y)

        chart_params, mthd_params = split_chart_params(**kwargs)

        out = StreamingChart(self, var_generator=var_generator,
                             steps=steps, interval=interval,
                             max_data=max_data, **chart_params)

        for item, params in zip(y, param_iter(mthd_params)):
            out.hbar(x, item, **params)

        return out

    def streaming_donut(self, x, y, steps=1e5, interval=1000,
                        max_data=100, var_generator=None, **kwargs):
        '''
        Display a streaming donut plot in a Jupyter notebook

        Notes
        -----
        Streaming figures are only supported in Jupyter notebooks.

        Parameters
        ----------
        x : string
            The slice categories
        y : string or list-of-strings
            The values of the slices
        steps : int, optional
            The maximum number of steps to iterate
        interval : int, optional
            The length of each step in milliseconds
        max_data : int, optional
            The maximum number of data points to display
        var_generator : callable, optional
            Function to call on new data to generate additional variables
        **kwargs : keyword arguments, optional
            Arguments to send to the StreamingChart object and donut method

        Examples
        --------
        Create the streaming figure.  Evaluating this in a Jupyter notebook
        will display the streaming plot.

        >>> fig = dataw.streaming_donut(x='time', y=['x', 'y', 'z'])
        >>> fig
        <esppy.plotting.StreamingChart at 0x7f98c0250e10>

        See Also
        --------
        :meth:`streaming_line`

        Returns
        -------
        :class:`StreamingChart`

        '''
        y = listify(y)

        chart_params, mthd_params = split_chart_params(**kwargs)

        out = StreamingChart(self, var_generator=var_generator,
                             steps=steps, interval=interval,
                             max_data=max_data, **chart_params)

        for item, params in zip(y, param_iter(mthd_params)):
            out.donut(x, item, **params)

        return out

    def streaming_pie(self, x, y, steps=1e5, interval=1000,
                      max_data=100, var_generator=None, **kwargs):
        '''
        Display a streaming pie plot in a Jupyter notebook

        Notes
        -----
        Streaming figures are only supported in Jupyter notebooks.

        Parameters
        ----------
        x : string
            The slice categories
        y : string or list-of-strings
            The values of the slices
        steps : int, optional
            The maximum number of steps to iterate
        interval : int, optional
            The length of each step in milliseconds
        max_data : int, optional
            The maximum number of data points to display
        var_generator : callable, optional
            Function to call on new data to generate additional variables
        **kwargs : keyword arguments, optional
            Arguments to send to the StreamingChart object and pie method

        Examples
        --------
        Create the streaming figure.  Evaluating this in a Jupyter notebook
        will display the streaming plot.

        >>> fig = dataw.streaming_pie(x='time', y=['x', 'y', 'z'])
        >>> fig
        <esppy.plotting.StreamingChart at 0x7f98c0250e10>

        See Also
        --------
        :meth:`streaming_line`

        Returns
        -------
        :class:`StreamingChart`

        '''
        y = listify(y)

        chart_params, mthd_params = split_chart_params(**kwargs)

        out = StreamingChart(self, var_generator=var_generator,
                             steps=steps, interval=interval,
                             max_data=max_data, **chart_params)

        for item, params in zip(y, param_iter(mthd_params)):
            out.pie(x, item, **params)

        return out

    def streaming_images(self, image_col, steps=1e5, interval=1000, size='contain',
                         var_generator=None, annotations=None, **kwargs):
        '''
        Display streaming images in a Jupyter notebook

        Notes
        -----
        Streaming images are only supported in Jupyter notebooks.

        Parameters
        ----------
        image_col : string
            The column which contains the image
        steps : int, optional
            The maximum number of steps to iterate
        interval : int, optional
            The length of each step in milliseconds
        size : string or int or tuple, optional
            The size of the displayed image.  It can take the following values:
                * 'contain' - to indicate that the image should be scaled to
                  fit inside the plot area
                * 'cover' - to indicate that the image should scale to cover
                  the entire plot area
                * int - to indicate a square with a pixel width of height of
                  the given integer
                * (int, int) - to indicate width and height values in pixels
        var_generator : callable, optional
            Function to call on new data to generate additional variables
        plot_width : int, optional
            The width of the plot in pixels
        plot_height : int, optional
            The height of the plot in pixels

        Returns
        -------
        :class:`StreamingImages`

        '''
        return StreamingImages(self.to_data_callback(extra=image_col,
                                                     var_generator=var_generator,
                                                     interval=interval),
                               steps=steps, interval=interval, size=size,
                               annotations=annotations, **kwargs)


class Window(BaseWindow, SplitterExpressionFeature,
             SplitterPluginFeature, FinalizedCallbackFeature,
             ConnectorsFeature):
    '''
    Standard ESP Window

    Parameters
    ----------
    name : string, optional
        The name of the window
    pubsub : bool, optional
        Publish/subscribe mode for the window. When the project-level value
        of pubsub is manual, true enables publishing and subscribing for
        the window and false disables it.
    description : string, optional
        Description of the window
    output_insert_only : bool, optional
        When true, prevents the window from passing non-insert events to
        other windows.
    collapse_updates : bool, optional
        When true, multiple update blocks are collapsed into a single
        update block
    pulse_interval : string, optional
        Output a canonical batch of updates at the specified interval
    exp_max_string : int, optional
        Specifies the maximum size of strings that the expression engine
        uses for the window. Default value is 1024.
    index_type : string, optional
        Index type for the window
        Valid values: 'rbtree', 'hash', 'ln_hash', 'cl_hash', 'fw_hash', 'empty'
    pubsub_index_type : string, optional
        Publish/subscribe index type.  Valid values are the same as for the
        `index_type` parameter.

    Attributes
    ----------
    connectors : list-of-Connectors
        The window connectors
    contquery : string
        The name of the continuous query the window is associated with
    data : pandas.DataFrame
        The cached rows of event data
    description : string
        A description of the window
    event_transformers : list
        A list of event transformer definitions
    finalized_callback : FinalizeCallback
        Shared library function for finalize callback
    project : string
        The name of the project the window is associated with
    splitter : SplitterExpression or SplitterPlugin
        Expression or plugin that directs events to one of n
        different output slots
    targets : set
        Set of target definitions
    url : string
        The URL of the window

    Returns
    -------
    :class:`Window`

    '''

    output_insert_only = attribute('output-insert-only', dtype='bool')
    collapse_updates = attribute('collapse-updates', dtype='bool')
    pulse_interval = attribute('pulse-interval', dtype='string')
    exp_max_string = attribute('exp-max-string', dtype='int')
    index_type = attribute('index', dtype='string', values=INDEX_TYPES)
    pubsub_index_type = attribute('pubsub-index', dtype='string', values=INDEX_TYPES)

    def __init__(self, name=None, pubsub=None, description=None,
                 output_insert_only=None, collapse_updates=None,
                 pulse_interval=None, exp_map_string=None, index_type=None,
                 pubsub_index_type=None, **kwargs):
        kwargs.update(**get_args(locals()))
        BaseWindow.__init__(self, **kwargs)



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

''' ESP Connectors '''

from __future__ import print_function, division, absolute_import, unicode_literals

import collections
import os
import re
import requests
import six
import xml.etree.ElementTree as ET
from six.moves import urllib
from .base import ESPObject, attribute
from .utils.rest import get_params
from .utils.data import get_project_data

StringParameter = collections.namedtuple('StringParameter', ['name', 'default'])
SelectParameter = collections.namedtuple('SelectParameter', ['name', 'default', 'values'])
BooleanParameter = collections.namedtuple('BoolParameter', ['name', 'default'])
NumericParameter = collections.namedtuple('NumberParameter', ['name', 'default'])


class ConnectorInfo(ESPObject):
    '''
    Base class for all ESP connectors

    Attributes
    ----------
    name : string
        Name of the algorithm
    required_params : dict
        Dictionary of required parameters
    optional_params : dict
        Dictionary of optional parameters

    Parameters
    ----------
    label : string
        Label for the algorithm

    '''

    label = attribute('label', dtype='string')
    type = attribute('type', dtype='string')

    def __init__(self, label, pubsub=None, type=None):
        ESPObject.__init__(self)
        self.label = label
        self.pubsub = pubsub
        self.type = type
        self.required_params = {}
        self.optional_params = {}

    @classmethod
    def from_xml(cls, data, session=None):
        '''
        Create algorithm object from XML definition

        Parameters
        ----------
        data : string
            XML string
        session : requests.Session, optional
            Session the connector is associated with

        Returns
        -------
        :class:`Connector`

        '''
        out = cls('')
        out.session = session

        out._set_attributes(data.attrib)

        for ptype, params in [('required', out.required_params),
                              ('optional', out.optional_params)]:
            for item in data.findall('./%s-parms/parm' % ptype):
                name = item.attrib['key']
                if item.find('./default') is not None:
                    default = item.find('./default').text
                    if item.find('./allowed-values') is not None:
                        if default in ['true', 'false']:
                            params[name] = BooleanParameter(name,
                                           (default == 'true') and True or False)
                        else:
                            params[name] = SelectParameter(name, default,
                                                           [x.text for x in
                                                            item.findall('.//allowed-value')])
                    elif re.match(r'^\d*\.\d+$', default):
                        params[name] = NumericParameter(name, float(default))
                    elif re.match(r'^\d+$', default):
                        params[name] = NumericParameter(name, int(default))
                    else:
                        params[name] = StringParameter(name, default)
                else:
                    params[name] = StringParameter(name, '')

        return out

    def __str__(self):
        attrs = ['label=%s' % repr(self.label)]
        if self.type:
            attrs.append('type=%s' % repr(self.type))
        if self.pubsub:
            attrs.append('pubsub=%s' % repr(self.pubsub))
        return '%s(%s)' % (type(self).__name__, ', '.join(attrs))

    def __repr__(self):
        return str(self)



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

''' Event Parsing Utilities '''

from __future__ import print_function, division, absolute_import, unicode_literals

import base64
import csv
import datetime
import decimal
import json
import numpy as np
import os
import pandas as pd
import re
import six
import sys
import xml.etree.ElementTree as ET
from six.moves import urllib
from ..base import ESPObject
from ..config import get_option

EPOCH = datetime.datetime(1970, 1, 1)


def str_to_float(value):
    ''' Convert value to float '''
    if isinstance(value, six.string_types):
        if value.strip() == '' or 'nan' in value.lower():
            return np.nan
    return np.float64(value)


def str_to_int32(value):
    ''' Convert value to float '''
    if isinstance(value, six.string_types):
        if value.strip() == '' or 'nan' in value.lower():
            return np.nan
    return np.int32(value)


def str_to_int64(value):
    ''' Convert value to float '''
    if isinstance(value, six.string_types):
        if value.strip() == '' or 'nan' in value.lower():
            return np.nan
    return np.int64(value)


def double_array(value):
    ''' Convert value to array of doubles '''
    if isinstance(value, six.string_types):
        if re.match(r'^\s*\[', value):
            out = [str_to_float(x)
                   for x in re.sub(r'[\[\]\s+]', r'', value).split(';')]
        else:
            out = [str_to_float(value)]
    else:
        out = [np.float64(value)]
    return np.array(out, dtype=np.float64)


def int32_array(value):
    ''' Convert value to array of int32s '''
    if isinstance(value, six.string_types):
        if re.match(r'^\s*\[', value):
            out = [str_to_int32(x)
                   for x in re.sub(r'[\[\]\s+]', r'', value).split(';')]
        else:
            out = [str_to_int32(value)]
    else:
        out = [np.int32(value)]
    return np.array(out, dtype=np.int32)


def int64_array(value):
    ''' Convert value to array of int32s '''
    if isinstance(value, six.string_types):
        if re.match(r'^\s*\[', value):
            out = [str_to_int64(x)
                   for x in re.sub(r'[\[\]\s+]', r'', value).split(';')]
        else:
            out = [str_to_int64(value)]
    else:
        out = [np.int64(value)]
    return np.array(out, dtype=np.int64)


ESP2PY_MAP = {
    'date': lambda x: EPOCH + datetime.timedelta(seconds=int(x)),
    'stamp': lambda x: EPOCH + datetime.timedelta(microseconds=int(x)),
    'double': str_to_float,
    'int64': lambda x: np.int64(decimal.Decimal(x)),
    'int32': lambda x: np.int32(decimal.Decimal(x)),
    'money': decimal.Decimal,
    'blob': base64.b64decode,
    'string': lambda x: hasattr(x, 'decode') and x.decode('utf-8') or x,
    'array(dbl)': double_array,
    'array(double)': double_array,
    'array(i32)': int32_array,
    'array(int32)': int32_array,
    'array(i64)': int64_array,
    'array(int64)': int64_array,
}

ESP2DF_TYPEMAP = {
    'date': datetime.datetime.now(),
    'stamp': datetime.datetime.now(),
    'double': np.float64(0),
    'int64': np.int64(0),
    'int32': np.int32(0),
    'money': decimal.Decimal(0),
    'blob': b'bytes',
    'string': u'string',
    'array(dbl)': pd.Series(dtype=np.float64),
    'array(double)': pd.Series(dtype=np.float64),
    'array(i32)': pd.Series(dtype=np.int32),
    'array(int32)': pd.Series(dtype=np.int32),
    'array(i64)': pd.Series(dtype=np.int64),
    'array(int64)': pd.Series(dtype=np.int64),
}


def get_dataframe(obj):
    '''
    Get an empty DataFrame that represents the Window schema

    Parameters
    ----------
    schema : Schema, optional
        The schema to use instead of the given window's schema

    Returns
    -------
    :class:`pandas.DataFrame`

    '''
    try:
        BaseWindow, Schema    # noqa: F821
    except:
        from ..schema import Schema
        from ..windows import BaseWindow

    if isinstance(obj, Schema):
        schema = obj
    elif isinstance(obj, BaseWindow) and obj.schema.fields:
        schema = obj.schema
    else:
        schema = get_schema(obj, obj.fullname)

    columns = []
    row = []
    index = []
    int32s = []
    int32_arrays = []
    for field in schema.fields.values():
        columns.append(field.name)
        row.append(ESP2DF_TYPEMAP[field.type])
        if field.type == 'int32':
            int32s.append(field.name)
        elif field.type in ['array(int32)', 'array(i32)']:
            int32s.append(field.name)
        if field.key:
            index.append(field.name)

    out = pd.DataFrame(columns=columns, data=[row])

    for name in int32s:
        out[name] = out[name].astype('int32')
    for name in int32_arrays:
        out[name] = out[name].apply(lambda x: x.astype('int32'),
                                    convert_dtype=False)

    if index:
        out = out.set_index(index)

    return out.iloc[0:0]


def get_schema(obj, window):
    ''' Retrieve the schema for the specified window '''
    try:
        get_window_class    # noqa: F821
    except:
        from ..windows import get_window_class

    if isinstance(window, six.string_types):
        path = window.replace('.', '/')
    else:
        if getattr(window, 'schema') and window.schema.fields:
            return window.schema.copy(deep=True)
        path = window.fullname.replace('.', '/')

    res = obj._get(urllib.parse.urljoin(obj.base_url, 'windows/%s' % path),
                   params=dict(schema='true'))

    for item in res.findall('./*'):
        try:
            wcls = get_window_class(item.tag)
        except KeyError:
            raise TypeError('Unknown window type: %s' % item.tag)
        return wcls.from_xml(item, session=obj.session).schema


def get_events(obj, data, format='xml', separator=None, single=False, server_info=None):
    '''
    Convert events to DataFrames

    Parameters
    ----------
    obj : ESPObject
        The calling object.  If this is a Schema, that schema is used
        for the events.  If it is a Window, the schema for tha window
        is used for the events.
    data : xml-string or ElementTree.Element
        The events to process
    format : string, optional
        The format of the events
    separator : string, optional
        The separator between each 'properties' events
    single : bool, optional
        Only return a single DataFrame rather than a dictionary.
        If there is more than one DataFrame, raise an exception.
    server_info : dict, optional
        Information about the server, for version-specific behaviors

    Returns
    -------
    dict of :class:`pandas.DataFrame`
        If single == False
    :class:`pandas.DataFrame`
        If single == True

    '''
    try:
        BaseWindow, Schema    # noqa: F821
    except:
        from ..schema import Schema
        from ..windows import BaseWindow

    server_info = server_info or {}

    if get_option('debug.events'):
        sys.stderr.write('%s\n' % data)

    if format.lower() == 'csv':
        return get_csv_events(obj, data)

    if format.lower() == 'json':
        return get_json_events(obj, data)

    if format.lower() == 'properties':
        try:
            return get_properties_events(obj, data, separator)
        except:
            import traceback
            traceback.print_exc()
            raise

    if isinstance(data, six.string_types):
        data = ET.fromstring(data)

    windows = dict()
    from . import xml
    for event in data.findall('./event'):

        wname = event.attrib.get('window', '')

        if wname not in windows:

            current = windows[wname] = dict(transformers={}, columns=[],
                                            index=[], events=[], dtypes=[])

            if isinstance(obj, Schema):
                schema = obj
            elif isinstance(obj, BaseWindow) and obj.fullname == wname.replace('/', '.') and obj.schema.fields:
                schema = obj.schema
            elif not wname:
                if isinstance(obj, BaseWindow):
                    schema = get_schema(obj, obj.fullname)
                else:
                    raise ValueError('Could not determine window schema')
            else:
                schema = get_schema(obj, wname)

            for field in schema.fields.values():
                current['transformers'][field.name] = ESP2PY_MAP.get(field.type,
                                                                     lambda x: x)
                current['columns'].append(field.name)
                current['dtypes'].append(field.type)
                if field.key:
                    current['index'].append(field.name)

        else:
            current = windows[wname]

        row = dict()
        for item in event.findall('./*'):
            row[item.tag] = current['transformers'].get(item.tag, lambda x: x)(item.text)
        current['events'].append(row)

    out = dict()
    for wname, window in windows.items():
        orig_wname = wname
        wname = wname.replace('/', '.')
        out[wname] = pd.DataFrame(window['events'])
        columns = [x for x in window['columns'] if x in out[wname].columns]
        out[wname] = out[wname][columns]
        for colname, dtype in zip(windows[orig_wname]['columns'],
                                  windows[orig_wname]['dtypes']):
            if dtype == 'int32':
                out[wname][colname] = out[wname][colname].astype('int32', copy=False)
        if window['index']:
            index = [x for x in window['index'] if x in out[wname].columns]
            if index:
                out[wname] = out[wname].set_index(index)

    if single:
        if len(out) == 1:
            return list(out.values())[0]
        elif not out:
            return get_dataframe(obj)
        raise ValueError('Output contains more than one value: %s' % out)

    return out


def get_csv_events(obj, data):
    '''
    Convert CSV events to DataFrames

    Parameters
    ----------
    obj : ESPObject
        The calling object.  If this is a Schema, that schema is used
        for the events.  If it is a Window, the schema for tha window
        is used for the events.
    data : csv-string
        The events to process

    Returns
    -------
    :class:`pandas.DataFrame`

    '''
    try:
        BaseWindow, Schema    # noqa: F821
    except:
        from ..schema import Schema
        from ..windows import BaseWindow

    if isinstance(obj, Schema):
        schema = obj
    elif isinstance(obj, BaseWindow):
        if obj.schema.fields:
            schema = obj.schema
        else:
            schema = get_schema(obj, obj)
    else:
        raise ValueError('Can not obtain window schema from given object')

    transformers = []
    columns = []
    index = []
    dtypes = []
    for fname, field in schema.fields.items():
        transformers.append(ESP2PY_MAP.get(field.type, lambda x: x))
        columns.append(fname)
        dtypes.append(field.type)
        if field.key:
            index.append(fname)

    rows = []
    for row in csv.reader(data.rstrip().split('\n')):
        row = list(row)[2:]
        for i, item in enumerate(row):
            row[i] = transformers[i](item)
        rows.append(row)

    out = pd.DataFrame(data=rows, columns=columns)
    out = out.set_index(index)
    for colname, dtype in zip(columns, dtypes):
        if dtype == 'int32':
            out[colname] = out[colname].astype('int32', copy=False)
    return out


def get_json_events(obj, data):
    '''
    Convert JSON events to DataFrames

    Parameters
    ----------
    obj : ESPObject
        The calling object.  If this is a Schema, that schema is used
        for the events.  If it is a Window, the schema for tha window
        is used for the events.
    data : json-string
        The events to process

    Returns
    -------
    :class:`pandas.DataFrame`

    '''
    try:
        BaseWindow, Schema    # noqa: F821
    except:
        from ..schema import Schema
        from ..windows import BaseWindow

    if isinstance(obj, Schema):
        schema = obj
    elif isinstance(obj, BaseWindow):
        if obj.schema.fields:
            schema = obj.schema
        else:
            schema = get_schema(obj, obj)
    else:
        raise ValueError('Can not obtain window schema from given object')

    transformers = {}
    columns = []
    index = []
    dtypes = []
    for fname, field in schema.fields.items():
        transformers[fname] = ESP2PY_MAP.get(field.type, lambda x: x)
        columns.append(fname)
        dtypes.append(field.type)
        if field.key:
            index.append(fname)

    rows = []
    for event in json.loads(data)['events']:
        event = event['event']
        row = []
        for col in columns:
            row.append(transformers[col](event[col]))
        rows.append(row)

    out = pd.DataFrame(data=rows, columns=columns)
    out = out.set_index(index)
    for colname, dtype in zip(columns, dtypes):
        if dtype == 'int32':
            out[colname] = out[colname].astype('int32', copy=False)
    return out


def get_properties_events(obj, data, separator=None):
    '''
    Convert properties events to DataFrames

    Parameters
    ----------
    obj : ESPObject
        The calling object.  If this is a Schema, that schema is used
        for the events.  If it is a Window, the schema for tha window
        is used for the events.
    data : json-string
        The events to process

    Returns
    -------
    :class:`pandas.DataFrame`

    '''
    try:
        BaseWindow, Schema    # noqa: F821
    except:
        from ..schema import Schema
        from ..windows import BaseWindow

    if separator is None:
        separator = '\n\n'

    if isinstance(obj, Schema):
        schema = obj
    elif isinstance(obj, BaseWindow):
        if obj.schema.fields:
            schema = obj.schema
        else:
            schema = get_schema(obj, obj)
    else:
        raise ValueError('Can not obtain window schema from given object: %s' % obj)

    transformers = {}
    columns = []
    index = []
    for fname, field in schema.fields.items():
        transformers[fname] = ESP2PY_MAP.get(field.type, lambda x: x)
        columns.append(fname)
        if field.key:
            index.append(fname)

    rows = []
    for event in [x for x in data.split(separator) if x.strip()]:
        row = []
        for i, col in enumerate(x for x in event.split('\n') if x.strip()):
            if i == 0 and col.startswith('opcode='):
                continue
            col, value = col.split('=', 1)
            row.append(transformers[col](value))
        rows.append(row)

    out = pd.DataFrame(data=rows, columns=columns)
    out = out.set_index(index)
    return out



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

''' ESP Database Connector '''

from __future__ import print_function, division, absolute_import, unicode_literals

import numbers
import re
import six
from .base import Connector, prop, map_properties
from ..utils import xml
from ..utils.data import gen_name


class DatabaseSubscriber(Connector):
    '''
    Subscribe to database events

    Parameters
    ----------
    connectstring : string
        Specifies the database DSN and user credentials in the
        format 'DSN=dsn;uid=userid;pwd=password;'
    desttablename : string
        Specifies the target table name
    snapshot : boolean, optional
        Specifies whether to send snapshot data
    configfilesection : string, optional
        Specifies the name of the section in the config file to parse for
        configuration parameters. Specify the value as [configfilesection].
    commitrows : int, optional
        Specifies the minimum number of output rows to buffer
    commitsecs : int, optional
        Specifies the maximum number of seconds to hold onto an incomplete
        commit buffer
    ignoresqlerrors : boolean, optional
        Enables the connector to continue to write Inserts, Updates, and
        Deletes to the database table despite an error in a previous Insert,
        Update, or Delete.
    maxcolbinding : int, optional
        Specifies the maximum supported width of string columns.
        The default value is 4096.
    pwdencrypted : boolean, optional
        Specifies that the pwd field in connectstring is encrypted
    rmretdel : boolean, optional
        Specifies to remove all delete events from event blocks received
        by a subscriber that were introduced by a window retention policy.

    Returns
    -------
    :class:`DatabaseSubscriber`

    '''
    connector_key = dict(cls='db', type='subscribe')

    property_defs = dict(
        connectstring=prop('connectstring', dtype='string', required=True),
        desttablename=prop('desttablename', dtype='string', required=True),
        snapshot=prop('snapshot', dtype='boolean', required=True),
        configfilesection=prop('configfilesection', dtype='string'),
        commitrows=prop('commitrows', dtype='int'),
        commitsecs=prop('commitsecs', dtype='int'),
        ignoresqlerrors=prop('ignoresqlerrors', dtype='boolean'),
        maxcolbinding=prop('maxcolbinding', dtype='int'),
        pwdencrypted=prop('pwdencrypted', dtype='boolean'),
        rmretdel=prop('rmretdel', dtype='boolean')
    )

    def __init__(self, connectstring=None, desttablename=None, name=None, is_active=None,
                 snapshot=None, configfilesection=None, commitrows=None,
                 commitsecs=None, ignoresqlerrors=None, maxcolbinding=None,
                 pwdencrypted=None, rmretdel=None):
        params = dict(**locals())
        params.pop('is_active')
        params.pop('self')
        name = params.pop('name')
        Connector.__init__(self, 'db', name=name, type='subscribe',
                           is_active=is_active, properties=params)

    @classmethod
    def from_parameters(cls, conncls, type=None, name=None, is_active=None,
                        properties=None):
        req, properties = map_properties(cls, properties,
                                         required=['connectstring',
                                                   'desttablename'],
                                         delete='type')
        return cls(req[0], req[1], name=name, is_active=is_active, **properties)


class DatabasePublisher(Connector):
    '''
    Subscribe events to a database

    Parameters
    ----------
    connectstring : string
        Specifies the database DSN and user credentials in the format
        'DSN=dsn;uid=userid;pwd=password;'
    blocksize : int, optional
        Specifies the number of events to include in a published event
        block. The default value is 1.
    configfilesection : string, optional
        Specifies the name of the section in the config file to parse for
        configuration parameters. Specify the value as [configfilesection].
    greenplumlogminer : boolean, optional
        Enables Greenplum log miner mode
    logminerdbname : string, optional
        Specifies the gpperfmon database that contains the queries_history
        table for Greenplum log miner mode. Use the following
        format: 'dd-mmm-yyy hh:mm:ss'.
    logminerschemaowner : string, optional
        Specifies the schema owner when using Oracle or Greenplum log
        miner mode.
    logminerstartdatetime : string, optional
        Specifies the start date time when using Oracle or Greenplum
        log miner mode.
    logminertablename : string, optional
        Specifies the table name when using Oracle or Greenplum log
        miner mode.
    maxcolbinding : int, optional
        Specifies the maximum supported width of string columns.
        The default value is 4096.
    maxevents : int, optional
        Specifies the maximum number of events to publish.
    oraclelogminer : boolean, optional
        Enables Oracle log miner mode
    publishwithupsert : boolean, optional
        Builds events with opcode=Upsert instead of Insert
    pwdencrypted : boolean, optional
        Specifies that the pwd field in connectstring is encrypted
    selectstatement : string, optional
        Specifies the SQL statement to be executed on the source database.
        Required when oraclelogminer and greenplumlogminer are not enabled.
    transactional : string, optional
        Sets the event block type to transactional. The default value is normal.

    Returns
    -------
    :class:`DatabasePublisher`

    '''
    connector_key = dict(cls='db', type='publish')

    property_defs = dict(
        connectstring=prop('connectstring', dtype='string', required=True),
        blocksize=prop('blocksize', dtype='int'),
        configfilesection=prop('configfilesection', dtype='string'),
        greenplumlogminer=prop('greenplumlogminer', dtype='boolean'),
        logminerdbname=prop('logminerdbname', dtype='string'),
        logminerschemaowner=prop('logminerschemaowner', dtype='string'),
        logminerstartdatetime=prop('logminerstartdatetime', dtype='string'),
        logminertablename=prop('logminertablename', dtype='string'),
        maxcolbinding=prop('maxcolbinding', dtype='int'),
        maxevents=prop('maxevents', dtype='int'),
        oraclelogminer=prop('oraclelogminer', dtype='boolean'),
        publishwithupsert=prop('publishwithupsert', dtype='boolean'),
        pwdencrypted=prop('pwdencrypted', dtype='boolean'),
        selectstatement=prop('selectstatement', dtype='string'),
        transactional=prop('transactional', dtype='string') 
    )

    def __init__(self, connectstring=None, name=None, is_active=None, blocksize=None,
                 configfilesection=None, greenplumlogminer=None,
                 logminerdbname=None, logminerschemaowner=None,
                 logminerstartdatetime=None, logminertablename=None,
                 maxcolbinding=None, maxevents=None, oraclelogminer=None,
                 publishwithupsert=None, pwdencrypted=None,
                 selectstatement=None, transactional=None):
        params = dict(**locals())
        params.pop('is_active')
        params.pop('self')
        name = params.pop('name')
        Connector.__init__(self, 'db', name=name, type='publish',
                           is_active=is_active, properties=params)

    @classmethod
    def from_parameters(cls, conncls, type=None, name=None, is_active=None,
                        properties=None):
        req, properties = map_properties(cls, properties,
                                         required=['connectstring'],
                                         delete='type')
        return cls(req[0], name=name, is_active=is_active, **properties)



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

# NOTE: This test requires a running ESP server.  You must use an ~/.authinfo
#       file to specify your username and password.  The ESP host and port must
#       be specified using the ESPHOST and ESPPORT environment variables.
#       A specific protocol ('http' or 'https') can be set using
#       the ESPPROTOCOL environment variable.

import datetime
import esppy
import os
import six
import sys
import time
import unittest
from esppy.utils.events import get_dataframe, get_schema, get_events
from . import utils as tm

USER, PASSWD = tm.get_user_pass()
HOST, PORT, PROTOCOL = tm.get_host_port_proto()
DATA_DIR = tm.get_data_dir()


class TestEventUtils(tm.TestCase):

    def setUp(self):
        self.s = esppy.ESP(HOST, PORT, USER, PASSWD, protocol=PROTOCOL)

        model_xml_path = os.path.join(DATA_DIR, 'model_sa.xml')
        name = 'ESPUnitTestProjectSA'

        with self.assertRaises(KeyError):
            self.s.get_project(name)

        self.project = self.s.install_project(model_xml_path)
        self.query = self.project.queries['contquery']

    def tearDown(self):
        self.s.delete_projects(filter="match(name, '^.*UnitTest.*$')")

    def test_get_dataframe(self):
        # From Window
        df = get_dataframe(self.query['w_data'])
        self.assertEqual(len(df), 0)
        self.assertEqual(list(df.columns), ['x_c', 'y_c'])
        self.assertEqual(list(df.index.names), ['id'])

        df = df.reset_index()
        self.assertEqual([str(x) for x in df.dtypes.tolist()],
                         ['int64', 'float64', 'float64'])

        # From Schema
        df = get_dataframe(self.query['w_data'].schema)
        self.assertEqual(len(df), 0)
        self.assertEqual(list(df.columns), ['x_c', 'y_c'])
        self.assertEqual(list(df.index.names), ['id'])

        df = df.reset_index()
        self.assertEqual([str(x) for x in df.dtypes.tolist()],
                         ['int64', 'float64', 'float64'])

        # From schema-less Window
        win = self.query['w_data']
        win.schema.fields.clear()
        df = get_dataframe(win)
        self.assertEqual(len(df), 0)
        self.assertEqual(list(df.columns), ['x_c', 'y_c'])
        self.assertEqual(list(df.index.names), ['id'])

        df = df.reset_index()
        self.assertEqual([str(x) for x in df.dtypes.tolist()],
                         ['int64', 'float64', 'float64'])

    def test_get_schema(self):
        sch = get_schema(self.s, 'ESPUnitTestProjectSA.contquery.w_data')
        self.assertEqual(sch.schema_string, 'id*:int64,x_c:double,y_c:double')
       
        sch = get_schema(self.s, self.query['w_data'])
        self.assertEqual(sch.schema_string, 'id*:int64,x_c:double,y_c:double')

        win = self.query['w_data']
        win.schema.fields.clear()
        sch = get_schema(self.s, win)
        self.assertEqual(sch.schema_string, 'id*:int64,x_c:double,y_c:double')

    def test_get_xml_events(self):
        data_path = os.path.join(DATA_DIR, 'data_sa.csv')
        w_data = self.query['w_data']
        sch = w_data.schema.copy(deep=True)

        # Window as obj
        def get_message(sock, message):
            out = get_events(w_data, message, format='xml', single=True) 
            self.assertEqual(list(out.index.names), ['id'])
            self.assertEqual(list(out.columns), ['x_c', 'y_c'])
            self.assertEqual(len(out), 1)

        sub = w_data.create_subscriber(mode='streaming', format='xml',
                                       on_message=get_message)
        self.assertTrue(sub.callbacks['on_message'] is get_message)
        sub.start()
        w_data.publish_events(data_path)
        time.sleep(3)
        sub.close()

        # Schema as obj
        def get_message(sock, message):
            out = get_events(sch, message, format='xml', single=True)
            self.assertEqual(list(out.index.names), ['id'])
            self.assertEqual(list(out.columns), ['x_c', 'y_c'])
            self.assertEqual(len(out), 1)

        sub = w_data.create_subscriber(mode='streaming', format='xml',
                                       on_message=get_message)
        self.assertTrue(sub.callbacks['on_message'] is get_message)
        sub.start()
        w_data.publish_events(data_path)
        time.sleep(3)
        sub.close()

        # Schema-less Window as obj
        w_data.schema.fields.clear()

        def get_message(sock, message):
            out = get_events(w_data, message, format='xml', single=True)
            self.assertEqual(list(out.index.names), ['id'])
            self.assertEqual(list(out.columns), ['x_c', 'y_c'])
            self.assertEqual(len(out), 1)

        sub = w_data.create_subscriber(mode='streaming', format='xml',
                                       on_message=get_message)
        self.assertTrue(sub.callbacks['on_message'] is get_message)
        sub.start()
        w_data.publish_events(data_path)
        time.sleep(3)
        sub.close()

    def test_get_csv_events(self):
        data_path = os.path.join(DATA_DIR, 'data_sa.csv')
        w_data = self.query['w_data']
        sch = w_data.schema.copy(deep=True)

        # Window as obj
        def get_message(sock, message):
            out = get_events(w_data, message, format='csv', single=True)
            self.assertEqual(list(out.index.names), ['id'])
            self.assertEqual(list(out.columns), ['x_c', 'y_c'])
            self.assertEqual(len(out), 1)

        sub = w_data.create_subscriber(mode='streaming', format='csv',
                                       on_message=get_message)
        self.assertTrue(sub.callbacks['on_message'] is get_message)
        sub.start()
        w_data.publish_events(data_path)
        time.sleep(3)
        sub.close()

        # Schema as obj
        def get_message(sock, message):
            out = get_events(sch, message, format='csv', single=True)
            self.assertEqual(list(out.index.names), ['id'])
            self.assertEqual(list(out.columns), ['x_c', 'y_c'])
            self.assertEqual(len(out), 1)

        sub = w_data.create_subscriber(mode='streaming', format='csv',
                                       on_message=get_message)
        self.assertTrue(sub.callbacks['on_message'] is get_message)
        sub.start()
        w_data.publish_events(data_path)
        time.sleep(3)
        sub.close()

        # Schema-less Window as obj
        w_data.schema.fields.clear()

        def get_message(sock, message):
            out = get_events(sch, message, format='csv', single=True)
            self.assertEqual(list(out.index.names), ['id'])
            self.assertEqual(list(out.columns), ['x_c', 'y_c'])
            self.assertEqual(len(out), 1)

        sub = w_data.create_subscriber(mode='streaming', format='csv',
                                       on_message=get_message)
        self.assertTrue(sub.callbacks['on_message'] is get_message)
        sub.start()
        w_data.publish_events(data_path)
        time.sleep(3)
        sub.close()

    def test_get_json_events(self):
        data_path = os.path.join(DATA_DIR, 'data_sa.csv')
        w_data = self.query['w_data']
        sch = w_data.schema.copy(deep=True)

        # Window as obj
        def get_message(sock, message):
            out = get_events(w_data, message, format='json', single=True)
            self.assertEqual(list(out.index.names), ['id'])
            self.assertEqual(list(out.columns), ['x_c', 'y_c'])
            self.assertEqual(len(out), 1)

        sub = w_data.create_subscriber(mode='streaming', format='json',
                                       on_message=get_message)
        self.assertTrue(sub.callbacks['on_message'] is get_message)
        sub.start()
        w_data.publish_events(data_path)
        time.sleep(3)
        sub.close()

        # Schema as obj
        def get_message(sock, message):
            out = get_events(sch, message, format='json', single=True)
            self.assertEqual(list(out.index.names), ['id'])
            self.assertEqual(list(out.columns), ['x_c', 'y_c'])
            self.assertEqual(len(out), 1)

        sub = w_data.create_subscriber(mode='streaming', format='json',
                                       on_message=get_message)
        self.assertTrue(sub.callbacks['on_message'] is get_message)
        sub.start()
        w_data.publish_events(data_path)
        time.sleep(3)
        sub.close()

        # Schema-less Window as obj
        w_data.schema.fields.clear()

        def get_message(sock, message):
            out = get_events(sch, message, format='json', single=True)
            self.assertEqual(list(out.index.names), ['id'])
            self.assertEqual(list(out.columns), ['x_c', 'y_c'])
            self.assertEqual(len(out), 1)

        sub = w_data.create_subscriber(mode='streaming', format='json',
                                       on_message=get_message)
        self.assertTrue(sub.callbacks['on_message'] is get_message)
        sub.start()
        w_data.publish_events(data_path)
        time.sleep(3)
        sub.close()

    def test_get_properties_events(self):
        data_path = os.path.join(DATA_DIR, 'data_sa.csv')
        w_data = self.query['w_data']
        sch = w_data.schema.copy(deep=True)

        # Window as obj
        def get_message(sock, message):
            out = get_events(w_data, message, format='properties', single=True)
            self.assertEqual(list(out.index.names), ['id'])
            self.assertEqual(list(out.columns), ['x_c', 'y_c'])
            self.assertEqual(len(out), 1)

        sub = w_data.create_subscriber(mode='streaming', format='properties',
                                       on_message=get_message)
        self.assertTrue(sub.callbacks['on_message'] is get_message)
        sub.start()
        w_data.publish_events(data_path)
        time.sleep(3)
        sub.close()

        # Schema as obj
        def get_message(sock, message):
            out = get_events(sch, message, format='properties', single=True)
            self.assertEqual(list(out.index.names), ['id'])
            self.assertEqual(list(out.columns), ['x_c', 'y_c'])
            self.assertEqual(len(out), 1)

        sub = w_data.create_subscriber(mode='streaming', format='properties',
                                       on_message=get_message)
        self.assertTrue(sub.callbacks['on_message'] is get_message)
        sub.start()
        w_data.publish_events(data_path)
        time.sleep(3)
        sub.close()

        # Schema-less Window as obj
        w_data.schema.fields.clear()

        # Schema-less Window as obj
        def get_message(sock, message):
            out = get_events(sch, message, format='properties', single=True)
            self.assertEqual(list(out.index.names), ['id'])
            self.assertEqual(list(out.columns), ['x_c', 'y_c'])
            self.assertEqual(len(out), 1)

        sub = w_data.create_subscriber(mode='streaming', format='properties',
                                       on_message=get_message)
        self.assertTrue(sub.callbacks['on_message'] is get_message)
        sub.start()
        w_data.publish_events(data_path)
        time.sleep(3)
        sub.close()


if __name__ == '__main__':
   tm.runtests()



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

from __future__ import print_function, division, absolute_import, unicode_literals

import six
from .base import Window, attribute
from .features import InitializeExpressionFeature, SchemaFeature
from .utils import get_args, ensure_element, connectors_to_end
from ..utils import xml


class FieldPlugin(object):
    '''
    Compute field plugin

    Parameters
    ----------
    plugin : string
       The shared library that contains the specified function
    function : string
       The specified function

    Returns
    -------
    :class:`FieldPlugin`

    '''

    def __init__(self, plugin, function):
        self.plugin = plugin
        self.function = function

    def copy(self, deep=False):
        return type(self)(self.plugin, self.function)


class ContextPlugin(object):
    '''
    Compute context plugin

    Parameters
    ----------
    name : string
        The shared library that contains the context–generation function
    function : string
        The function that, when called, returns a new derived context
        for the window’s handler routines.

    Returns
    -------
    :class:`ContextPlugin`

    '''

    def __init__(self, name, function):
        self.name = name
        self.function = function

    def copy(self, deep=False):
        return type(self)(self.name, self.function)


class ComputeWindow(Window, InitializeExpressionFeature, SchemaFeature):
    '''
    Compute window

    Parameters
    ----------
    name : string, optional
        The name of the window
    schema : Schema
        The schema of the window
    pubsub : bool, optional
        Publish/subscribe mode for the window. When the project-level value
        of pubsub is manual, true enables publishing and subscribing for
        the window and false disables it.
    description : string, optional
        Description of the window
    output_insert_only : bool, optional
        When true, prevents the window from passing non-insert events to
        other windows.
    collapse_updates : bool, optional
        When true, multiple update blocks are collapsed into a single
        update block
    pulse_interval : string, optional
        Output a canonical batch of updates at the specified interval
    exp_max_string : int, optional
        Specifies the maximum size of strings that the expression engine
        uses for the window. Default value is 1024.
    index_type : string, optional
        Index type for the window
    pubsub_index_type : string, optional
        Publish/subscribe index type.  Valid values are the same as the
        `index_type` parameter.

    Attributes
    ----------
    context_plugin : ContextPlugin
        Function that returns context for use in plugins
    field_expressions : list-of-strings
        Field expressions
    field_plugins : list-of-FieldPlugins
        Field plugins

    Returns
    -------
    :class:`ComputeWindow`

    '''

    window_type = 'compute'

    def __init__(self, name=None, schema=None, pubsub=None, description=None,
                 output_insert_only=None, collapse_updates=None,
                 pulse_interval=None, exp_max_string=None, index_type=None,
                 pubsub_index_type=None):
        Window.__init__(self, **get_args(locals()))
        self.context_plugin = None
        self.field_expressions = []
        self.field_plugins = []

    def copy(self, deep=False):
        out = Window.copy(self, deep=deep)
        out.context_plugin = self.context_plugin.copy(deep=deep)
        if deep:
            out.field_expressions = [x.copy(deep=deep) for x in self.field_expressions]
            out.field_plugins = [x.copy(deep=deep) for x in self.field_plugins]
        else: 
            out.field_expressions = list(self.field_expressions)
            out.field_plugins = list(self.field_plugins)
        return out

    @classmethod
    def from_element(cls, data, session=None):
        '''
        Convert XML / Element to object

        Parameters
        ----------
        data : xml-string or Element
            The element to convert
        session : Session, optional
            The requests.Session object

        Returns
        -------
        :class:`ComputeWindow`

        '''
        data = ensure_element(data)
        out = super(ComputeWindow, cls).from_element(data, session=session)

        for item in data.findall('./context-plugin'):
            out.context_plugin = ContextPlugin(item.attrib['name'],
                                               item.attrib['function'])

        for item in data.findall('./output/field-expr'):
            out.field_expressions.append(item.text)

        for item in data.findall('./output/field-plug'):
            out.field_plugins.append(FieldPlugin(item.attrib['plugin'],
                                                 item.attrib['function']))

        return out

    from_xml = from_element

    def to_element(self, query=None):
        '''
        Convert object to Element

        Returns
        -------
        :class:`ElementTree.Element`

        '''
        out = Window.to_element(self, query=query)

        if self.context_plugin is not None:
            xml.add_elem(out, 'context-plugin',
                         attrib=dict(name=self.context_plugin.name,
                                     function=self.context_plugin.function))

        schema = out.find('./schema')
        if schema is not None:
            out.remove(schema)
            out.append(schema)

        output = xml.add_elem(out, 'output')

        for item in self.field_expressions:
            xml.add_elem(output, 'field-expr', text_content=item)

        for item in self.field_plugins:
            xml.add_elem(output, 'field-plug',
                         attrib=dict(plugin=item.plugin, function=item.function))

        connectors_to_end(out)

        return out

    def set_context_plugin(self, name, function):
        '''
        Set a context plugin

        Parameters
        ----------
        name : string
            The shared library that contains the context–generation function
        function : string
            The function that, when called, returns a new derived context
            for the window’s handler routines.

        '''
        self.context_plugin = ContextPlugin(name, function)

    def add_field_expressions(self, *expr):
        '''
        Add new field expressions

        Parameters
        ----------
        *expr : one-or-more-strings
            The expressions to add

        '''
        for exp in expr:
            self.field_expressions.append(exp)

    add_field_expression = add_field_expressions

    add_field_expr = add_field_expressions

    add_field_exprs = add_field_expressions

    def add_field_plugin(self, plugin, function):
        '''
        Add a field plugin

        Parameters
        ----------
        plugin : string
            The name of the plugin
        function : string or list-of-strings
            The name(s) of the function

        '''
        if isinstance(function, six.string_types):
            function = [function]
        for func in function:
            self.field_plugins.append(FieldPlugin(plugin, func))



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

# NOTE: This test requires a running ESP server.  You must use an ~/.authinfo
#       file to specify your username and password.  The ESP host and port must
#       be specified using the ESPHOST and ESPPORT environment variables.
#       A specific protocol ('http' or 'https') can be set using
#       the ESPPROTOCOL environment variable.

import copy
import datetime
import os
import six
import esppy
import sys
import unittest
from esppy.schema import SchemaField, Schema
from esppy.utils import xml
from . import utils as tm

USER, PASSWD = tm.get_user_pass()
HOST, PORT, PROTOCOL = tm.get_host_port_proto()
DATA_DIR = tm.get_data_dir()


class TestSchemaField(tm.TestCase):

    def setUp(self):
        self.s = esppy.ESP(HOST, PORT, USER, PASSWD, protocol=PROTOCOL)

        model_xml_path = os.path.join(DATA_DIR, 'model_sa.xml')
        name = 'ESPUnitTestProjectSA'

        with self.assertRaises(KeyError):
            self.s.get_project(name)

        self.project = self.s.install_project(model_xml_path)
        self.query = self.project.queries['contquery']

    def tearDown(self):
        self.s.delete_projects(filter="match(name, '^.*UnitTest.*$')")

    def test_init(self):
        field = SchemaField('a', 'int64', key='true')
        self.assertEqual(field.name, 'a')
        self.assertEqual(field.type, 'int64')
        self.assertTrue(field.key is True)

        field = SchemaField('a', 'int64', key='false')
        self.assertEqual(field.name, 'a')
        self.assertEqual(field.type, 'int64')
        self.assertTrue(field.key is False)

        field = SchemaField('a', 'int64', key=1)
        self.assertEqual(field.name, 'a')
        self.assertEqual(field.type, 'int64')
        self.assertTrue(field.key is True)

        field = SchemaField('a', 'int64', key=0)
        self.assertEqual(field.name, 'a')
        self.assertEqual(field.type, 'int64')
        self.assertTrue(field.key is False)

    def test_copy(self):
        field = SchemaField('foo', 'int64', key='true')

        # Use method
        fcopy = field.copy()
        self.assertTrue(field is not fcopy)
        self.assertEqual(field.name, fcopy.name) 
        self.assertEqual(field.type, fcopy.type) 
        self.assertEqual(field.key, fcopy.key) 

        # Use method with deep=True
        fcopy = field.copy(deep=True)
        self.assertTrue(field is not fcopy)
        self.assertEqual(field.name, fcopy.name) 
        self.assertEqual(field.type, fcopy.type) 
        self.assertEqual(field.key, fcopy.key) 

        # Use copy.copy
        fcopy = copy.copy(field)
        self.assertTrue(field is not fcopy)
        self.assertEqual(field.name, fcopy.name) 
        self.assertEqual(field.type, fcopy.type) 
        self.assertEqual(field.key, fcopy.key) 

        # Use copy.deepcopy
        fcopy = copy.deepcopy(field)
        self.assertTrue(field is not fcopy)
        self.assertEqual(field.name, fcopy.name) 
        self.assertEqual(field.type, fcopy.type) 
        self.assertEqual(field.key, fcopy.key) 

    def test_from_xml(self):
        field = SchemaField.from_xml('<field name="foo" type="int64" />')
        self.assertEqual(field.name, 'foo')
        self.assertEqual(field.type, 'int64')
        self.assertEqual(field.key, False)

        field = SchemaField.from_xml('<field name="foo" type="int64" key="true" />')
        self.assertEqual(field.name, 'foo')
        self.assertEqual(field.type, 'int64')
        self.assertEqual(field.key, True)

        field = SchemaField.from_xml(xml.new_elem('field', attrib=dict(name='foo',
                                                                       type='int32',
                                                                       key=True)))
        self.assertEqual(field.name, 'foo')
        self.assertEqual(field.type, 'int32')
        self.assertEqual(field.key, True)

    def test_to_element(self):
        field = SchemaField.from_xml('<field name="foo" type="int64" />')
        field = field.to_element()
        self.assertEqual(field.tag, 'field')
        self.assertEqual(field.attrib, {'type': 'int64', 'name': 'foo', 'key': 'false'})

    def test_to_xml(self):
        field = SchemaField.from_xml('<field name="foo" type="int64" />')
        field = field.to_xml()
        self.assertEqual(field, '<field key="false" name="foo" type="int64" />')

    def test_str(self):
        field = SchemaField('foo', 'int64', key='true')
        self.assertEqual(str(field), 'foo*:int64')

        field = SchemaField('foo', 'int64', key='false')
        self.assertEqual(str(field), 'foo:int64')

    def test_repr(self):
        field = SchemaField('foo', 'int64', key='true')
        self.assertEqual(repr(field), 'foo*:int64')

        field = SchemaField('foo', 'int64', key='false')
        self.assertEqual(repr(field), 'foo:int64')


class TestSchema(tm.TestCase):

    def setUp(self):
        self.s = esppy.ESP(HOST, PORT, USER, PASSWD, protocol=PROTOCOL)

        model_xml_path = os.path.join(DATA_DIR, 'model_sa.xml')
        name = 'ESPUnitTestProjectSA'

        with self.assertRaises(KeyError):
            self.s.get_project(name)

        self.project = self.s.install_project(model_xml_path)
        self.query = self.project.queries['contquery']

    def tearDown(self):
        self.s.delete_projects(filter="match(name, '^.*UnitTest.*$')")

    def test_init(self):
        sch = Schema(SchemaField('a', 'int64'),
                     SchemaField('b', 'double', key=True),
                     SchemaField('c', 'int32'))
        self.assertEqual(list(sch.keys()), ['a', 'b', 'c'])
        self.assertEqual([x.type for x in sch.values()],
                         ['int64', 'double', 'int32']) 
        self.assertEqual([x.key for x in sch.values()],
                         [False, True, False])

    def test_copy(self):
        win = self.project['contquery']['w_data']

        # Use the copy method
        scopy = win.schema.copy()
        self.assertEqual(list(scopy.keys()), ['id', 'x_c', 'y_c'])
        self.assertEqual([x.type for x in scopy.values()],
                         ['int64', 'double', 'double'])
        self.assertEqual([x.key for x in scopy.values()],
                         [True, False, False])
        self.assertTrue(scopy['id'] is win.schema['id'])
        self.assertTrue(scopy['x_c'] is win.schema['x_c'])
        self.assertTrue(scopy['y_c'] is win.schema['y_c'])

        # Use the copy method with deep=True
        scopy = win.schema.copy(deep=True)
        self.assertEqual(list(scopy.keys()), ['id', 'x_c', 'y_c'])
        self.assertEqual([x.type for x in scopy.values()],
                         ['int64', 'double', 'double'])
        self.assertEqual([x.key for x in scopy.values()],
                         [True, False, False])
        self.assertTrue(scopy['id'] is not win.schema['id'])
        self.assertTrue(scopy['x_c'] is not win.schema['x_c'])
        self.assertTrue(scopy['y_c'] is not win.schema['y_c'])

        # Use copy.copy
        scopy = copy.copy(win.schema)
        self.assertEqual(list(scopy.keys()), ['id', 'x_c', 'y_c'])
        self.assertEqual([x.type for x in scopy.values()],
                         ['int64', 'double', 'double'])
        self.assertEqual([x.key for x in scopy.values()],
                         [True, False, False])
        self.assertTrue(scopy['id'] is win.schema['id'])
        self.assertTrue(scopy['x_c'] is win.schema['x_c'])
        self.assertTrue(scopy['y_c'] is win.schema['y_c'])

        # Use copy.deepcopy
        scopy = copy.deepcopy(win.schema)
        self.assertEqual(list(scopy.keys()), ['id', 'x_c', 'y_c'])
        self.assertEqual([x.type for x in scopy.values()],
                         ['int64', 'double', 'double'])
        self.assertEqual([x.key for x in scopy.values()],
                         [True, False, False])
        self.assertTrue(scopy['id'] is not win.schema['id'])
        self.assertTrue(scopy['x_c'] is not win.schema['x_c'])
        self.assertTrue(scopy['y_c'] is not win.schema['y_c'])

    def test_schema_string(self):
        sch = Schema(SchemaField('foo', 'int64'),
                     SchemaField('bar', 'double', key=True),
                     SchemaField('baz', 'int32'))
        
        self.assertEqual(len(sch), 3)
        self.assertEqual(sch.schema_string, 'foo:int64,bar*:double,baz:int32')

        sch.schema_string = 'one*:int64,two:int64,three:double,four:int64'

        self.assertEqual(sch.schema_string, 
                         'one*:int64,two:int64,three:double,four:int64')
        self.assertEqual(len(sch), 4)
        self.assertEqual(sch.fields['one'].name, 'one')
        self.assertEqual(sch.fields['one'].type, 'int64')
        self.assertEqual(sch.fields['one'].key, True)
        self.assertEqual(sch.fields['two'].name, 'two')
        self.assertEqual(sch.fields['two'].type, 'int64')
        self.assertEqual(sch.fields['two'].key, False)
        self.assertEqual(sch.fields['three'].name, 'three')
        self.assertEqual(sch.fields['three'].type, 'double')
        self.assertEqual(sch.fields['three'].key, False)
        self.assertEqual(sch.fields['four'].name, 'four')
        self.assertEqual(sch.fields['four'].type, 'int64')
        self.assertEqual(sch.fields['four'].key, False)

        sch.schema_string = 'one*:int64,two'

        self.assertEqual(len(sch), 2)
        self.assertEqual(sch.fields['one'].name, 'one')
        self.assertEqual(sch.fields['one'].type, 'int64')
        self.assertEqual(sch.fields['one'].key, True)
        self.assertEqual(sch.fields['two'].name, 'two')
        self.assertEqual(sch.fields['two'].type, 'inherit')
        self.assertEqual(sch.fields['two'].key, False)

    def test_from_xml(self):
        sch = Schema.from_xml('<schema><fields>'
                              '<field name="foo" type="int64" key="true"/>'
                              '<field name="bar" type="int32"/>'
                              '<field name="baz" type="int32"/>'
                              '</fields></schema>')

        self.assertEqual(len(sch), 3)
        self.assertEqual(sch.fields['foo'].name, 'foo')
        self.assertEqual(sch.fields['foo'].type, 'int64')
        self.assertEqual(sch.fields['foo'].key, True)
        self.assertEqual(sch.fields['bar'].name, 'bar')
        self.assertEqual(sch.fields['bar'].type, 'int32')
        self.assertEqual(sch.fields['bar'].key, False)
        self.assertEqual(sch.fields['baz'].name, 'baz')
        self.assertEqual(sch.fields['baz'].type, 'int32')
        self.assertEqual(sch.fields['baz'].key, False)

    def test_to_element(self):
        sch = Schema.from_xml('<schema><fields>'
                              '<field name="foo" type="int64" key="true"/>'
                              '<field name="bar" type="int32"/>'
                              '<field name="baz" type="int32"/>'
                              '</fields></schema>')
        elem = sch.to_element()
        self.assertEqual(elem.tag, 'schema')
        for fields in elem:
            for i, field in enumerate(fields):
                if i == 0:
                    self.assertEqual(field.tag, 'field')
                    self.assertEqual(field.attrib, dict(name='foo', type='int64', key='true'))
                elif i == 1:
                    self.assertEqual(field.tag, 'field')
                    self.assertEqual(field.attrib, dict(name='bar', type='int32', key='false'))
                elif i == 2:
                    self.assertEqual(field.tag, 'field')
                    self.assertEqual(field.attrib, dict(name='baz', type='int32', key='false'))
                else:
                    raise ValueError('Too many items')

    def test_to_xml(self):
        sch = Schema.from_xml('<schema><fields>' + \
                              '<field name="foo" type="int64" key="true" />' + \
                              '<field name="bar" type="int32" key="false" />' + \
                              '<field name="baz" type="int32" key="false" />' + \
                              '</fields></schema>')
        self.assertEqual(sch.to_xml(), '<schema><fields>'
                                       '<field key="true" name="foo" type="int64" />'
                                       '<field key="false" name="bar" type="int32" />'
                                       '<field key="false" name="baz" type="int32" />'
                                       '</fields></schema>')

    def test_setitem(self):
        sch = Schema()
        sch['foo'] = 'double', True
        sch['bar'] = 'int32'
        sch['baz*'] = 'int64'
        self.assertEqual(sch.schema_string, 'foo*:double,bar:int32,baz*:int64')
        self.assertEqual(list(sch.keys()), ['foo', 'bar', 'baz'])
        self.assertEqual([x.name for x in sch.values()], ['foo', 'bar', 'baz'])
        self.assertEqual([x.type for x in sch.values()], ['double', 'int32', 'int64'])
        self.assertEqual([x.key for x in sch.values()], [True, False, True])

    def test_update_field(self):
        sch = Schema()
        sch['foo'] = 'double', True
        sch['bar'] = 'int32'
        self.assertEqual(sch.schema_string, 'foo*:double,bar:int32')

        sch['foo'].type = 'int64'
        self.assertEqual(sch.schema_string, 'foo*:int64,bar:int32')

        sch['bar'].key = True
        self.assertEqual(sch.schema_string, 'foo*:int64,bar*:int32')

    def test_delitem(self):
        sch = Schema()
        sch['foo'] = 'double', True
        sch['bar'] = 'int32'

        self.assertEqual(list(sch.fields.keys()), ['foo', 'bar'])

        del sch['foo']

        self.assertEqual(list(sch.fields.keys()), ['bar'])

    def test_str(self):
        sch = Schema()
        sch.schema_string = 'foo*:double,bar:int64'
        self.assertEqual(str(sch), 'foo*:double,bar:int64')

    def test_repr(self):
        sch = Schema()
        sch.schema_string = 'foo*:double,bar:int64'
        self.assertEqual(repr(sch), 'foo*:double,bar:int64')

    def test_mapping_methods(self):
        sch = Schema()
        sch.schema_string = 'foo*:double,bar:int64,baz:int32'
        
        self.assertEqual(sch['foo'].name, 'foo')
        self.assertEqual(sch['foo'].type, 'double')
        self.assertEqual(sch['foo'].key, True)

        sch['foo'] = SchemaField('foo', 'int64', key=False)
        self.assertEqual(sch['foo'].name, 'foo')
        self.assertEqual(sch['foo'].type, 'int64')
        self.assertEqual(sch['foo'].key, False)

        self.assertEqual(sch.schema_string, 'foo:int64,bar:int64,baz:int32')
        del sch['foo']
        self.assertEqual(sch.schema_string, 'bar:int64,baz:int32')

        self.assertEqual(len(list(sch)), 2)
        for i, item in enumerate(sch):
            if i == 0:
                self.assertEqual(item, 'bar')
            elif i == 1:
                self.assertEqual(item, 'baz')

        self.assertEqual(len(sch), 2) 
 

if __name__ == '__main__':
   tm.runtests()



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

# NOTE: This test requires a running ESP server.  You must use an ~/.authinfo
#       file to specify your username and password.  The ESP host and port must
#       be specified using the ESPHOST and ESPPORT environment variables.
#       A specific protocol ('http' or 'https') can be set using
#       the ESPPROTOCOL environment variable.

import os
import re
import six
import esppy
import unittest
from .. import utils as tm
from ..utils import elements_equal, normalize_xml
from ...utils import xml

USER, PASSWD = tm.get_user_pass()
HOST, PORT, PROTOCOL = tm.get_host_port_proto()
DATA_DIR = tm.get_data_dir()


class TestWindowFeatures(tm.TestCase):

    def setUp(self):
        self.s = esppy.ESP(HOST, PORT, USER, PASSWD, protocol=PROTOCOL)

    def tearDown(self):
        self.s.delete_projects(filter="match(name, '^.*UnitTest.*$')")

    def _test_model_file(self, model):
        model_file = os.path.join(DATA_DIR, model)
        model_xml = normalize_xml(open(model_file, 'r').read())

        proj1 = self.s.Project.from_xml(model_xml)
        proj1_xml = normalize_xml(proj1.to_xml(pretty=True))

        proj2 = self.s.load_project(model_file, start=False)
        proj2_xml = normalize_xml(proj2.to_xml(pretty=True))

        elements_equal(model_xml, proj1_xml)
        elements_equal(model_xml, proj2_xml)

    def test_splitter(self):
        self._test_model_file('splitter_model.xml')

    def test_filter_exp(self):
        self._test_model_file('filter_exp_model.xml')

    def test_retention(self):
        self._test_model_file('retention_model.xml')

    def test_slot_exp(self):
        self._test_model_file('slot_exp_model.xml')

    def test_aggregate_exp(self):
        self._test_model_file('aggregate_exp_model.xml')

    @unittest.skip('Need way to install plugin')
    def test_aggregate_func(self):
        self._test_model_file('aggregate_func_model.xml')

    def test_aggregate_last(self):
        self._test_model_file('aggregate_last_model.xml')
    
    def test_aggregate_userdef(self):
        self._test_model_file('aggregate_userdef_model.xml')
    
    def test_analytics_dbscan(self):
        self._test_model_file('analytics_dbscan_model.xml')
    
    def test_analytics_kmeans(self):
        self._test_model_file('analytics_kmeans_model.xml')
    
    def test_broker_surv(self):
        self._test_model_file('broker_surv_model.xml')

    def test_calculate_img_proc(self):
        self._test_model_file('calculate_img_proc_model.xml')
    
    @unittest.skip('Need way to install plugin')
    def test_compute_context(self):
        self._test_model_file('compute_context_model.xml')
    
    def test_compute_exp_udf(self):
        self._test_model_file('compute_exp_udf_model.xml')
    
    def test_compute_exp(self):
        self._test_model_file('compute_exp_model.xml')
    
    @unittest.skip('Need way to install plugin')
    def test_compute_func(self):
        self._test_model_file('compute_func_model.xml')

    def test_compute_udf(self):
        self._test_model_file('compute_udf_model.xml')

    def test_copy_with_slots(self):
        self._test_model_file('copy_with_slots_model.xml')

    def test_db_connector_publisher(self):
        self._test_model_file('db_connector_publisher_model.xml')

    def test_db_connector_subscriber(self):
        self._test_model_file('db_connector_subscriber_model.xml')

    @unittest.skip('Need way to install plugin')
    def test_filter_func(self):
        self._test_model_file('filter_func_model.xml')

    def test_filter_on_flag(self):
        self._test_model_file('filter_on_flag_model.xml')

    def test_filter_on_opcode(self):
        self._test_model_file('filter_on_opcode_model.xml')

    @unittest.skip('Update to 5.2 specs')
    def test_ds2_procedural(self):
        self._test_model_file('ds2_procedural_model.xml')

    @unittest.skip('Crashes server')
    def test_ds_procedural(self):
        self._test_model_file('ds_procedural_model.xml')

    @unittest.skip('Update to 5.2 specs')
    def test_mas_composite(self):
        self._test_model_file('mas_composite_model.xml')

    @unittest.skip('Update to 5.2 specs')
    def test_mas_ds2_procedural(self):
        self._test_model_file('mas_ds2_procedural_model.xml')

    @unittest.skip('Update to 5.2 specs')
    def test_mas_python_procedural(self):
        self._test_model_file('mas_python_procedural_model.xml')

    def test_fs_adapter_publish(self):
        self._test_model_file('fs_adapter_publish_model.xml')

    @unittest.skip('Invalid resource')
    def test_functional_event_loop(self):
        self._test_model_file('functional_event_loop_model.xml')

    def test_geofence2(self):
        self._test_model_file('geofence2_model.xml')

    def test_geofence_proximity_analysis(self):
        self._test_model_file('geofence_proximity_analysis_model.xml')

    def test_geofence(self):
        self._test_model_file('geofence_model.xml')

    def test_join_exp(self):
        self._test_model_file('join_exp_model.xml')

    def test_join_select(self):
        self._test_model_file('join_select_model.xml')

    def test_log_config(self):
        self._test_model_file('log_config_model.xml')

    def test_orders(self):
        self._test_model_file('orders_model.xml')

    @unittest.skip('Need way to install plugin')
    def test_pattern2(self):
        self._test_model_file('pattern2_model.xml')

    def test_pattern_empty_index(self):
        self._test_model_file('pattern_empty_index_model.xml')

    def test_pattern(self):
        self._test_model_file('pattern_model.xml')

    @unittest.skip('Need way to install plugin')
    def test_procedural(self):
        self._test_model_file('procedural_model.xml')

    def test_regex(self):
        self._test_model_file('regex_model.xml')

    def test_splitter_initexp(self):
        self._test_model_file('splitter_initexp_model.xml')

    def test_splitter_udf(self):
        self._test_model_file('splitter_udf_model.xml')

    @unittest.skip('Needs mco file')
    def test_text_category(self):
        self._test_model_file('text_category_model.xml')

    @unittest.skip('Needs sam file')
    def test_text_sentiment(self):
        self._test_model_file('text_sentiment_model.xml')

    def test_textcat(self):
        self._test_model_file('textcat_model.xml')

    def test_trades(self):
        self._test_model_file('trades_model.xml')

    def test_twitter_adapter(self):
        self._test_model_file('twitter_adapter_model.xml')

    def test_union(self):
        self._test_model_file('union_model.xml')

    def test_url_connector_news(self):
        self._test_model_file('url_connector_news_model.xml')

    def test_url_connector_weather(self):
        self._test_model_file('url_connector_weather_model.xml')

    def test_vwap(self):
        self._test_model_file('vwap_model.xml')

    @unittest.skip('Update to 5.2 specs')
    def test_wind_turbine(self):
        self._test_model_file('wind_turbine_model.xml')

    def test_ws_connector_json(self):
        self._test_model_file('ws_connector_json_model.xml')

    def test_ws_connector_json2(self):
        self._test_model_file('ws_connector_json2_model.xml')

    def test_xml_connector_publisher(self):
        self._test_model_file('xml_connector_publisher_model.xml')

    def test_xml_connector_subscriber(self):
        self._test_model_file('xml_connector_subscriber_model.xml')


if __name__ == '__main__':
   tm.runtests()



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

''' ESP Routers '''

from __future__ import print_function, division, absolute_import, unicode_literals

import collections
import os
import re
import requests
import six
from six.moves import urllib
from .base import ESPObject
from .utils import xml
from .utils.rest import get_params
from .utils.data import get_project_data, gen_name


class Engine(object):
    '''
    ESP Engine Configuration

    Parameters
    ----------
    name : string
        Name of the ESP engine
    host : string
        Hostname of the ESP server
    port : string
        Port number of the ESP server
    auth_token : string, optional
        Auth token
    auth_token_url : string, optional
        Auth token URL

    Attributes
    ----------
    name : string
        Name of the ESP engine
    host : string
        Hostname of the ESP server
    port : string
        Port number of the ESP server
    auth_token : string
        Auth token
    auth_token_url : string
        Auth token URL
    router : string
        The name of the router the engine definiton belongs to

    Returns
    -------
    :class:`Engine`

    '''

    def __init__(self, host, port, name=None, auth_token=None, auth_token_url=None):
        self.name = name or gen_name(prefix='eng_')
        self.host = host
        self.port = int(port)
        self.auth_token = auth_token
        self.auth_token_url = auth_token_url
        self.router = None

    def to_element(self):
        eng = xml.new_elem('esp-engine',
                           attrib=dict(name=self.name, host=self.host,
                                       port=self.port))
        if self.auth_token is not None:
            xml.add_elem(eng, 'auth_token', self.auth_token)
        if self.auth_token_url is not None:
            xml.add_elem(eng, 'auth_token_url', self.auth_token_url)
        return eng

    def to_xml(self, pretty=False):
        return xml.to_xml(self.to_element(), pretty=pretty)


class PublishDestination(ESPObject):
    '''
    Router Publish Destination

    Parameters
    ----------
    target : string
        The target path in the form "<engine>.<project>.<contquery>.<window>"
    name : string, optional
        The name of the destination
    opcode : string, optional
        The opcode to apply to each event
    filter_func : string, optional
        The function used to filter the events that are published. The function
        is run for each event. Its Boolean value is evaluated to determine
        whether to publish the event to the appropriate target Source window.
    event_fields_init : dict, optional
        Container for functions to be run in order to initialize variables.
        These variables can be used in any subsequent functions.
    event_fields : dict, optional
        Container for functions to be run in order to add or modify event fields.

    Attributes
    ----------
    target : string
        The target path in the form "<engine>.<project>.<contquery>.<window>"
    name : string
        The name of the destination
    opcode : string
        The opcode to apply to each event
    filter_func : string
        The function used to filter the events that are published. The function
        is run for each event. Its Boolean value is evaluated to determine
        whether to publish the event to the appropriate target Source window.
    event_fields_init : dict
        Container for functions to be run in order to initialize variables.
        These variables can be used in any subsequent functions.
    event_fields : dict
        Container for functions to be run in order to add or modify event fields.
    engine_func : string
        Function used to resolve the target engine. It must resolve to the name
        of one of the engines that are defined in the router context.
    project_func : string
        Function used to resolve the target project. It must resolve to the name
        of a project in the engine that is resolved in engine_func.
    contquery_func : string
        Function used to resolve the target continuous query. It must resolve
        to the name of a continuous query in the project that is resolved in
        project-func.
    window_func : string
        Function used to resolve the target Source window. It must resolve to the
        name of a source window in the continuous query that is resolved in
        contquery_func.
    router : string
        The name of the router the destination belongs to

    Returns
    -------
    :class:`PublishDestination`

    '''

    def __init__(self, target, name=None, opcode=None, filter_func=None,
                 event_fields_init=None, event_fields=None):
        ESPObject.__init__(self)
        self.name = name or gen_name(prefix='pd_')
        self.opcode = opcode
        self.filter_func = filter_func
        self.event_fields_init = dict(event_fields_init or {})
        self.event_fields = dict(event_fields or {})
        self.target = target
        self.router = None

    @property
    def target(self):
        '''
        Target path

        Returns
        -------
        string

        '''
        return '.'.join(['%s' % x for x in [self.engine_func,
                                            self.project_func,
                                            self.contquery_func,
                                            self.window_func]])

    @target.setter
    def target(self, value):
        '''
        Set the target path

        Parameters
        ----------
        value : string
            The target path in the form "<engine>.<project>.<contquery>.<window>"

        '''
        if value.count('.') < 3:
            raise ValueError('Target does not contain enough levels')
        value = list(value.split('.'))
        self.engine_func = value[0]
        self.project_func = value[1]
        self.contquery_func = value[2]
        self.window_func = value[3]

    def initialize(self):
        ''' Initialize the destination '''
        self._put(urllib.parse.urljoin(self.base_url,
                                       'routerDestinations/%s/%s/state' %
                                       (self.router, self.name)),
                  params=get_params(value='initialized'))

    def to_element(self):
        '''
        Convert destination to Element definition

        Returns
        -------
        :class:`ElementTree.Element`

        '''
        dest = xml.new_elem('publish-destination',
                            attrib=dict(name=self.name,
                                        opcode=self.opcode))
        if self.filter_func is not None:
            xml.add_elem(dest, 'filter-func',
                         text_content=self.filter_func)

        tgt = xml.add_elem(dest, 'publish-target')
        if self.engine_func is not None:
            xml.add_elem(tgt, 'engine-func', text_content=self.engine_func)
        if self.project_func is not None:
            xml.add_elem(tgt, 'project-func', text_content=self.project_func)
        if self.contquery_func is not None:
            xml.add_elem(tgt, 'contquery-func', text_content=self.contquery_func)
        if self.window_func is not None:
            xml.add_elem(tgt, 'window-func', text_content=self.window_func)

        if self.event_fields_init or self.event_fields:
            efields = xml.add_elem(dest, 'event-fields')
            if self.event_fields_init:
                init = xml.add_elem(efields, 'init')
                for key, value in six.iteritems(self.event_fields_init):
                    xml.add_elem(init, 'value', attrib=dict(name=key),
                                 text_content=value)
            if self.event_fields:
                flds = xml.add_elem(efields, 'fields')
                for key, value in six.iteritems(self.event_fields):
                    xml.add_elem(flds, 'field', attrib=dict(name=key),
                                 text_content=value)

        return dest

    def to_xml(self, pretty=False):
        '''
        Convert destination to XML definition

        Parameters
        ----------
        pretty : bool, optional
            Should whitespace be added for readability?

        Returns
        -------
        string

        '''
        return xml.to_xml(self.to_element(), pretty=pretty)


class WriterDestination(ESPObject):
    '''
    Route Writer Destination

    Parameters
    ----------
    file_func : string
        Function used to resolve the name of the file into which the events
        are written.
    format : string
        Format of event output. Valid values are XML, JSON, and CSV.
        The default is XML.
    name : string, optional
        The name of the destination
    dateformat : string, optional
        The format for datetime strings in the data

    Attributes
    ----------
    file_func : string
        Function used to resolve the name of the file into which the events
        are written.
    format : string
        Format of event output. Valid values are XML, JSON, and CSV.
        The default is XML.
    name : string
        The name of the destination
    dateformat : string
        The format for datetime strings in the data
    router : string
        The name of the router the destination belongs to

    Returns
    -------
    :class:`WriterDestination`

    '''

    def __init__(self, file_func, format, name=None, dateformat='%Y%m%dT%H:%M:%S.%f'):
        ESPObject.__init__(self)
        self.name = name or gen_name(prefix='wd_')
        self.format = format
        self.dateformat = dateformat
        self.file_func = file_func
        self.router = None

    def initialize(self):
        ''' Initialize the destination '''
        self._put(urllib.parse.urljoin(self.base_url,
                                       'routerDestinations/%s/%s/state' %
                                       (self.router, self.name)),
                  params=get_params(value='initialized'))

    def to_element(self):
        '''
        Convert the destination to an Element definition

        Returns
        -------
        :class:`WriterDestination`

        '''
        dest = xml.new_elem('writer-destination',
                            attrib=dict(name=self.name,
                                        format=self.format,
                                        dateformat=self.dateformat))
        xml.add_elem(dest, 'file-func', text_content=self.file_func)
        return dest

    def to_xml(self, pretty=False):
        '''
        Convert the destination to an XML definition

        Parameters
        ----------
        pretty : bool, optional
            Should whitespace be added for readability?

        Returns
        -------
        :class:`ElementTree.Element`

        '''
        return xml.to_xml(self.to_element(), pretty=pretty)


class Route(object):
    '''
    Route

    Parameters
    ----------
    route : string
        The route path in the form "<engine>.<project>.<contquery>.<window>.<type>"
        where <type> is optional.
    to : string
        Comma-separated list of destinations that receive all of the events coming
        in from the subscriptions contained within the route.
    name : string, optional
        The name of the route
    snapshot : bool, optional
        Specify a value of true when you want to grab a snapshot of the initial
        window contents.

    Attributes
    ----------
    route : string
        The route path in the form "<engine>.<project>.<contquery>.<window>.<type>"
        where <type> is optional.
    to : string
        Comma-separated list of destinations that receive all of the events coming
        in from the subscriptions contained within the route.
    name : string, optional
        The name of the route
    snapshot : bool, optional
        Specify a value of true when you want to grab a snapshot of the initial
        window contents.
    engine_expr : string
        Regular expression used to resolve the engine(s) to which the route should
        subscribe. If it is not specified, the route subscribes to all engines.
    project_expr : string
        Regular expression used to resolve the project(s) to which the route should
        subscribe. If it is not specified, the route subscribes to all projects.
    contquery_expr : string
        Regular expression used to resolve the continuous queries to which the route
        should subscribe. If it is not specified, the route subscribes to all
        continuous queries.
    window_expr : string
        Regular expression used to resolve the window(s) to which the route
        should subscribe. If it is not specified, the route subscribes to
        all windows.
    type_expr : string
        Regular expression used to resolve the window type(s) to which the
        route should subscribe. If it is not specified, the route subscribes
        to all window types.
    router : string
        The name of the router the route belongs to

    Returns
    -------
    :class:`Route`

    '''

    def __init__(self, route, to, name=None, snapshot=False):
        self.name = name or gen_name('r_')
        self.to = to
        self.snapshot = snapshot
        self.route = route
        self.router = None

    @property
    def route(self):
        '''
        Route path

        Returns
        -------
        string

        '''
        route = '.'.join(['%s' % x for x in [self.engine_expr,
                                             self.project_expr,
                                             self.contquery_expr,
                                             self.window_expr,
                                             self.type_expr]])
        while route.endswith('.None'):
            route = route[:-5]
        return route

    @route.setter
    def route(self, value):
        '''
        Set route path

        Parameters
        ----------
        value : string
            The route path in the form "<engine>.<project>.<contquery>.<window>.<type>"
            where <type> is optional.

        '''
        if value.count('.') < 3:
            raise ValueError('Route does not contain enough levels')
        value = list(value.split('.')) + [None]
        self.engine_expr = value[0]
        self.project_expr = value[1]
        self.contquery_expr = value[2]
        self.window_expr = value[3]
        self.type_expr = value[4]

    def to_element(self):
        '''
        Convert Route to an Element definition

        Returns
        -------
        :class:`ElementTree.Element`

        '''
        rte = xml.new_elem('esp-route',
                           attrib=dict(name=self.name, to=self.to,
                                       snapshot=self.snapshot))

        if self.engine_expr is not None:
            xml.add_elem(rte, 'engine-expr', text_content=self.engine_expr)
        if self.project_expr is not None:
                    xml.add_elem(rte, 'project-expr', text_content=self.project_expr)
        if self.contquery_expr is not None:
            xml.add_elem(rte, 'contquery-expr', text_content=self.contquery_expr)
        if self.window_expr is not None:
            xml.add_elem(rte, 'window-expr', text_content=self.window_expr)
        if self.type_expr is not None:
            xml.add_elem(rte, 'type-expr', text_content=self.type_expr)

        return rte

    def to_xml(self, pretty=False):
        '''
        Convert Route to an XML definition

        Parameters
        ----------
        pretty : bool, optional
            Should whitespace be added for readability?

        Returns
        -------
        string

        '''
        return xml.to_xml(self.to_element(), pretty=pretty)


class Router(ESPObject):
    '''
    Router definition

    Parameters
    ----------
    name : string
        Name of the router

    Attributes
    ----------
    name : string
        Name of the router
    engines : dict-of-Engines
        The ESP engines in the router definition
    destinations : dict-of-PublishDestinations/WriterDestinations
        The destinations for the router
    routes : dict-of-Routes
        The routes defined in the Router

    '''

    def __init__(self, name=None):
        ESPObject.__init__(self)
        self.name = name or gen_name(prefix='r_')
        self.engines = {}
        self.destinations = {}
        self.routes = {}

    @classmethod
    def from_xml(cls, data, session=None):
        '''
        Create router from XML definition

        Parameters
        ----------
        data : xml-string or ElementTree.Element
            XML router definition
        session : requests.Session, optional
            Session that the router is associated with

        Returns
        -------
        :class:`Router`

        '''
        out = cls()
        out.session = session

        if isinstance(data, six.string_types):
            data = xml.from_xml(data)

        if data.tag != 'esp-router':
            data = data.find('.//esp-router')
            if data is None:
                raise ValueError('No router definition was found')

        out.name = data.attrib['name']

        for eng in data.findall('./esp-engines/esp-engine'):
            args = dict(eng.attrib)
            for item in eng.findall('./auth_token'):
                args['auth_token'] = item.text
            for item in eng.findall('./auth_token_url'):
                args['auth_token_url'] = item.text
            eng = Engine(**args)
            eng.session = session
            eng.router = out.name
            out.engines[eng.name] = eng

        for pdest in data.findall('./esp-destinations/publish-destination'):
            path = ['None', 'None', 'None', 'None']
            for item in pdest.findall('./publish-target/engine-func'):
                path[0] = item.text
            for item in pdest.findall('./publish-target/project-func'):
                path[1] = item.text
            for item in pdest.findall('./publish-target/contquery-func'):
                path[2] = item.text
            for item in pdest.findall('./publish-target/window-func'):
                path[3] = item.text

            filter_func = None
            for item in pdest.findall('./filter-func'):
                filter_func = item.text

            einit = {}
            for evt in data.findall('./event-fields/init/value'):
                name = evt.attrib.get('name', gen_name(prefix='ei_'))
                einit[name] = evt.text

            efields = {}
            for evt in data.findall('./event-fields/fields/field'):
                name = evt.attrib.get('name', gen_name(prefix='ef_'))
                efields[name] = evt.text

            dst = PublishDestination('.'.join(path), filter_func=filter_func,
                                     event_fields_init=einit, event_fields=efields,
                                     **pdest.attrib)
            dst.session = session
            dst.router = out.name
            out.destinations[dst.name] = dst

        for wdest in data.findall('./esp-destinations/writer-destination'):
            args = dict(wdest.attrib)
            for func in wdest.findall('./file-func'):
                args['file_func'] = func.text
            dst = WriterDestination(**args)
            dst.session = session
            dst.router = out.name
            out.destinations[dst.name] = dst

        for rte in data.findall('./esp-routes/esp-route'):
            path = ['None', 'None', 'None', 'None']
            for item in rte.findall('./engine-expr'):
                path[0] = item.text
            for item in rte.findall('./project-expr'):
                path[1] = item.text
            for item in rte.findall('./contquery-expr'):
                path[2] = item.text
            for item in rte.findall('./window-expr'):
                path[3] = item.text
            for item in rte.findall('./type-expr'):
                path.append(item.text)
            path = '.'.join(path)

            route = Route(path, rte.attrib['to'],
                          name=rte.attrib.get('name'),
                          snapshot=rte.attrib.get('snapshot',
                                                  'f').lower().startswith('t'))
            route.session = session
            route.router = out.name
            out.routes[route.name] = route

        return out

    from_element = from_xml

    def to_element(self):
        '''
        Convert Router to Element definition

        Returns
        -------
        :class:`ElementTree.Element`

        '''
        out = xml.new_elem('esp-router', attrib=dict(name=self.name))

        if self.engines:
            engs = xml.add_elem(out, 'esp-engines')
            for item in sorted(self.engines.values(), key=lambda x: x.name):
                xml.add_elem(engs, item.to_element())

        if self.destinations:
            dests = xml.add_elem(out, 'esp-destinations')
            for item in sorted(self.destinations.values(), key=lambda x: x.name):
                xml.add_elem(dests, item.to_element())

        if self.routes:
            routes = xml.add_elem(out, 'esp-routes')
            for item in sorted(self.routes.values(), key=lambda x: x.name):
                xml.add_elem(routes, item.to_element())

        return out

    def to_xml(self, pretty=False):
        '''
        Convert Router to XML definition

        Parameters
        ----------
        pretty : bool, optional
            Should whitespace be added for readability?

        Returns
        -------
        string

        '''
        return xml.to_xml(self.to_element(), pretty=pretty)

    def add_engine(self, host, port, name=None, auth_token=None,
                   auth_token_url=None):
        '''
        Add a new router engine

        Parameters
        ----------
        name : string
            Name of the router engine
        host : string
            Hostname of the server
        port : int
            Port number of the server
        auth_token : string, optional
            Auth token
        auth_token_url : string, optional
            URL to auth token

        Returns
        -------
        :class:`Engine`

        '''
        eng = Engine(host, port, name=name, auth_token=auth_token,
                     auth_token_url=auth_token_url)
        self.engines[eng.name] = eng
        return eng

    def add_publish_destination(self, target, name=None, opcode=None, filter_func=None,
                                event_fields_init=None, event_fields=None):
        '''
        Add a new router publish destination

        Parameters
        ----------
        target : string
            The target path in the form "<engine>.<project>.<contquery>.<window>"
        name : string, optional
            Name of the router destination
        opcode : string, optional
            Opcode for each event
        filter_func : string, optional
            The function used to filter the events that are published. The
            function is run for each event. Its Boolean value is evaluated
            to determine whether to publish the event to the appropriate
            target Source window.
        event_fields_init : dict, optional
            Container for functions to be run in order to initialize
            variables. These variables can be used in any subsequent
            functions.
        event_fields : dict, optional
            Container for functions to be run in order to add or modify
            event fields.

        '''
        dst = PublishDestination(target, name=name, opcode=opcode, filter_func=filter_func,
                                 event_fields_init=event_fields_init,
                                 event_fields=event_fields)
        dst.session = self.session
        dst.router = self.name
        self.destinations[dst.name] = dst

    def add_writer_destination(self, file_func, format, name=None,
                               dateformat='%Y%m%dT%H:%M:%S.%f'):
        '''
        Add a new router writer destination

        Parameters
        ----------
        file_func : string
            Function used to resolve the name of the file into which the
            events are written.
        format : string
            Format of event output. Valid values are XML, JSON, and CSV.
            The default is XML.
        name : string, optional
            Name of the router destination
        dateformat : string, optional
            Format of date strings

        Returns
        -------
        :class:`WriterDestination`

        '''
        dst = WriterDestination(file_func, format, name=name, dateformat=dateformat)
        dst.session = self.session
        dst.router = self.name
        self.destinations[dst.name] = dst
        return dst

    def initialize_destination(self, name):
        '''
        Initalize router destination

        Parameters
        ----------
        name : string
            Name of the destination to initialize

        '''
        self.destinations[name].initialize()

    def add_route(self, route, to, name=None, snapshot=False):
        '''
        Add a new route

        Parameters
        ----------
        route : string
            The route path in the form
            "<engine>.<project>.<contquery>.<window>.<type>", where <type>
            is optional.
        to : string
            Comma-separated list of destinations that receive all of the
            events coming in from the subscriptions contained within the route.
        name : string, optional
            Name of the route to create
        snapshot : bool, optional
            Specify a value of true when you want to grab a snapshot of
            the initial window contents.

        Returns
        -------
        :class:`Route`

        '''
        rte = Route(route, to, name=name, snapshot=snapshot)
        rte.session = self.session
        rte.router = self.name
        self.routes[rte.name] = rte
        return rte

    def save(self, overwrite=True):
        '''
        Save the router definition to the server

        Parameters
        ----------
        overwrite : bool, optional
            Should an existing router of the same name be overwritten?

        '''
        self._put(urllib.parse.urljoin(self.base_url,
                                       'routers/%s' % self.name),
                  params=get_params(overwrite=overwrite),
                  data=get_project_data(self))

    def delete(self):
        ''' Delete the router '''
        self._delete(urllib.parse.urljoin(self.base_url,
                                          'routers/%s' % self.name))
